/*******************************************************************************
  This is the driver for the ST MAC 10/100/1000 on-chip Ethernet controllers.
  ST Ethernet IPs are built around a Synopsys IP Core.

  Copyright (C) 2007-2009  STMicroelectronics Ltd

  This program is free software; you can redistribute it and/or modify it
  under the terms and conditions of the GNU General Public License,
  version 2, as published by the Free Software Foundation.

  This program is distributed in the hope it will be useful, but WITHOUT
  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
  more details.

  You should have received a copy of the GNU General Public License along with
  this program; if not, write to the Free Software Foundation, Inc.,
  51 Franklin St - Fifth Floor, Boston, MA 02110-1301 USA.

  The full GNU General Public License is included in this distribution in
  the file called "COPYING".

  Author: Giuseppe Cavallaro <peppe.cavallaro@st.com>

  Documentation available at:
	http://www.stlinux.com
  Support available at:
	https://bugzilla.stlinux.com/
*******************************************************************************/

#include <linux/module.h>
#include <linux/init.h>
#include <linux/kernel.h>
#include <linux/interrupt.h>
#include <linux/etherdevice.h>
#include <linux/platform_device.h>
#include <linux/ip.h>
#include <linux/tcp.h>
#include <linux/skbuff.h>
#include <linux/ethtool.h>
#include <linux/if_ether.h>
#include <linux/crc32.h>
#include <linux/mii.h>
#include <linux/phy.h>
#include <linux/if_vlan.h>
#include <linux/dma-mapping.h>
#include <linux/slab.h>
#include <mach/gpio.h>
#include <linux/proc_fs.h>
#include "stmmac.h"
#include "gs_misc.h"

#define STMMAC_RESOURCE_NAME	"stmmaceth"
#define PHY_RESOURCE_NAME	"stmmacphy"

#undef STMMAC_DEBUG
/*#define STMMAC_DEBUG*/
#ifdef STMMAC_DEBUG
#define DBG(nlevel, klevel, fmt, args...) \
		((void)(netif_msg_##nlevel(priv) && \
		printk(KERN_##klevel fmt, ## args)))
#else
#define DBG(nlevel, klevel, fmt, args...) do { } while (0)
#endif

#undef STMMAC_RX_DEBUG
/*#define STMMAC_RX_DEBUG*/
#ifdef STMMAC_RX_DEBUG
#define RX_DBG(fmt, args...)  printk(fmt, ## args)
#else
#define RX_DBG(fmt, args...)  do { } while (0)
#endif

#undef STMMAC_XMIT_DEBUG
/*#define STMMAC_XMIT_DEBUG*/
#ifdef STMMAC_TX_DEBUG
#define TX_DBG(fmt, args...)  printk(fmt, ## args)
#else
#define TX_DBG(fmt, args...)  do { } while (0)
#endif

#define STMMAC_ALIGN(x)	L1_CACHE_ALIGN(x)
#define JUMBO_LEN	9000

/* Module parameters */
#define TX_TIMEO 5000 /* default 5 seconds */
static int watchdog = TX_TIMEO;
module_param(watchdog, int, S_IRUGO | S_IWUSR);
MODULE_PARM_DESC(watchdog, "Transmit timeout in milliseconds");

static int debug = -1;		/* -1: default, 0: no output, 16:  all */
module_param(debug, int, S_IRUGO | S_IWUSR);
MODULE_PARM_DESC(debug, "Message Level (0: no output, 16: all)");

static int phyaddr = -1;
module_param(phyaddr, int, S_IRUGO);
MODULE_PARM_DESC(phyaddr, "Physical device address");

#define DMA_TX_SIZE 256
static int dma_txsize = DMA_TX_SIZE;
module_param(dma_txsize, int, S_IRUGO | S_IWUSR);
MODULE_PARM_DESC(dma_txsize, "Number of descriptors in the TX list");

#define DMA_RX_SIZE 256
static int dma_rxsize = DMA_RX_SIZE;
module_param(dma_rxsize, int, S_IRUGO | S_IWUSR);
MODULE_PARM_DESC(dma_rxsize, "Number of descriptors in the RX list");

static int flow_ctrl = FLOW_OFF;
module_param(flow_ctrl, int, S_IRUGO | S_IWUSR);
MODULE_PARM_DESC(flow_ctrl, "Flow control ability [on/off]");

static int pause = PAUSE_TIME;
module_param(pause, int, S_IRUGO | S_IWUSR);
MODULE_PARM_DESC(pause, "Flow Control Pause Time");

#define TC_DEFAULT 64
static int tc = TC_DEFAULT;
module_param(tc, int, S_IRUGO | S_IWUSR);
MODULE_PARM_DESC(tc, "DMA threshold control value");

#define RX_NO_COALESCE	1	/* Always interrupt on completion */
#define TX_NO_COALESCE	-1	/* No moderation by default */

/* Pay attention to tune this parameter; take care of both
 * hardware capability and network stabitily/performance impact.
 * Many tests showed that ~4ms latency seems to be good enough. */
#ifdef CONFIG_STMMAC_TIMER
#define DEFAULT_PERIODIC_RATE	256
static int tmrate = DEFAULT_PERIODIC_RATE;
module_param(tmrate, int, S_IRUGO | S_IWUSR);
MODULE_PARM_DESC(tmrate, "External timer freq. (default: 256Hz)");
#endif

#define DMA_BUFFER_SIZE	BUF_SIZE_2KiB
static int buf_sz = DMA_BUFFER_SIZE;
module_param(buf_sz, int, S_IRUGO | S_IWUSR);
MODULE_PARM_DESC(buf_sz, "DMA buffer size");

static const u32 default_msg_level = (NETIF_MSG_DRV | NETIF_MSG_PROBE |
				      NETIF_MSG_LINK | NETIF_MSG_IFUP |
				      NETIF_MSG_IFDOWN | NETIF_MSG_TIMER);

static irqreturn_t stmmac_interrupt(int irq, void *dev_id);
static netdev_tx_t stmmac_xmit(struct sk_buff *skb, struct net_device *dev);

#ifdef CONFIG_RTL8363_CPUPORT
#ifndef USE_VIRTUAL_ETH
#error "you must defined USE_VIRTUAL_ETH by CONFIG_GXE_STMMAC_VETH first"
#endif
#define ETH_P_RTL8899	0x8899          /* Realtek Remote Control Protocol (RRCP) Type, */

struct rrcp_hdr {
 //   __be16		h_rrcp_proto;
    __be16		h_rrcp_proto_reason;  
    __be16		h_rrcp_fid_prior;
    __be16		h_rrcp_portmask;
};

#endif

#ifdef USE_VIRTUAL_ETH
#define REAL_NDEV_NAME "stmmac%d"
#else
#define REAL_NDEV_NAME "eth%d"
#endif
#ifdef USE_VIRTUAL_ETH
    #ifndef CONFIG_RTL8363_CPUPORT
    #define LAN_VLANID   101
    #define WAN_VLANID  102
    #define SVIP_VLANID  1001
    #endif
#define NUM_OF_VDEV 3
#define LAN_LED_NAME "led_lan"
#define WAN_LED_NAME "led_wan"
#define LAN_LED_RJ45_NAME "led_lan_rj45"
#define WAN_LED_RJ45_NAME "led_wan_rj45"

struct net_device *real_ndev = NULL;

enum PORT_ID {
    GXE_LAN_PORT = 1,
    GXE_WAN_PORT = 2,
    GXE_DMW96_PORT = 5,
    GXE_SVIP_PORT = 6,
};

struct g_virt_netdev {
   struct net_device *ndev;
#ifdef CONFIG_RTL8363_CPUPORT
    u16 port;
#else
   u16 vlan_id;
#endif
   int opened;
   int link_state;
} virt_ndev[NUM_OF_VDEV] = {
#ifdef CONFIG_RTL8363_CPUPORT
    {NULL, GXE_SVIP_PORT, 0 , 0}, /*SVIP*/
    {NULL, GXE_LAN_PORT, 0, 0},  /*LAN*/
    {NULL, GXE_WAN_PORT, 0, 0},  /*WAN*/
#else
    {NULL, SVIP_VLANID, 0 , 0}, /*SVIP*/
    {NULL, LAN_VLANID, 0, 0},  /*LAN*/
    {NULL, WAN_VLANID, 0, 0},  /*WAN*/
#endif
};

struct stmmac_mdio_priv g_mdio_priv = {
    .phy_inused = 0,
    .has_nphys = 0,
};

int g_route_switch_mode = 0;
int g_link_num = 0;

extern struct gsboard_gpio * __init gs_get_board_gpio(int *count);

static int get_free_phydev (void)
{
    int ii;
    for(ii=0;ii<g_mdio_priv.has_nphys;ii++){
        if(g_mdio_priv.phy_inused & (1<<ii))
            continue;
        g_mdio_priv.phy_inused |= (1<<ii);
        return g_mdio_priv.phy_addr[ii];
    }
    return -1;
}

#ifdef ETH_P_RTL8899
static struct net_device *find_virt_ndev(u16 port)
{
    int ii;
    
    for(ii=0;ii<NUM_OF_VDEV;ii++){
        if(port == virt_ndev[ii].port)
            return virt_ndev[ii].ndev;
    }  
    return NULL;
}

static int find_vndev_port(struct net_device *dev)
{
    int ii;
    
    if(!dev) 
        return -1;
    
    for(ii=0;ii<NUM_OF_VDEV;ii++){
       if(virt_ndev[ii].ndev == dev)
           return  virt_ndev[ii].port;
    }
    return -1;
}

#else //esle--#ifdef ETH_P_RTL8899
static struct net_device *find_virt_ndev(u16 vlan_id)
{
    int ii;
    
    for(ii=0;ii<NUM_OF_VDEV;ii++){
        if(vlan_id == virt_ndev[ii].vlan_id)
            return virt_ndev[ii].ndev;
    }  
    return NULL;
}
static int find_vndev_vid(struct net_device *dev)
{
    int ii;
    
    if(!dev) 
        return -1;
    
    for(ii=0;ii<NUM_OF_VDEV;ii++){
       if(virt_ndev[ii].ndev == dev)
           return  virt_ndev[ii].vlan_id;
    }
    return -1;
}
#endif  //end--#ifdef ETH_P_RTL8899

#endif //end--#ifdef USE_VIRTUAL_ETH

/**
 * stmmac_verify_args - verify the driver parameters.
 * Description: it verifies if some wrong parameter is passed to the driver.
 * Note that wrong parameters are replaced with the default values.
 */
static void stmmac_verify_args(void)
{
	if (unlikely(watchdog < 0))
		watchdog = TX_TIMEO;
	if (unlikely(dma_rxsize < 0))
		dma_rxsize = DMA_RX_SIZE;
	if (unlikely(dma_txsize < 0))
		dma_txsize = DMA_TX_SIZE;
	if (unlikely((buf_sz < DMA_BUFFER_SIZE) || (buf_sz > BUF_SIZE_16KiB)))
		buf_sz = DMA_BUFFER_SIZE;
	if (unlikely(flow_ctrl > 1))
		flow_ctrl = FLOW_AUTO;
	else if (likely(flow_ctrl < 0))
		flow_ctrl = FLOW_OFF;
	if (unlikely((pause < 0) || (pause > 0xffff)))
		pause = PAUSE_TIME;
}

#if defined(STMMAC_XMIT_DEBUG) || defined(STMMAC_RX_DEBUG)
static void print_pkt(unsigned char *buf, int len)
{
	int j;
	pr_info("len = %d byte, buf addr: 0x%p", len, buf);
	for (j = 0; j < len; j++) {
		if ((j % 16) == 0)
			pr_info("\n %03x:", j);
		pr_info(" %02x", buf[j]);
	}
	pr_info("\n");
}
#endif

/* minimum number of free TX descriptors required to wake up TX process */
#define STMMAC_TX_THRESH(x)	(x->dma_tx_size/4)

static inline u32 stmmac_tx_avail(struct stmmac_priv *priv)
{
	return priv->dirty_tx + priv->dma_tx_size - priv->cur_tx - 1;
}

/* On some ST platforms, some HW system configuraton registers have to be
 * set according to the link speed negotiated.
 */
static inline void stmmac_hw_fix_mac_speed(struct stmmac_priv *priv)
{
	struct phy_device *phydev = priv->phydev;

	if (likely(priv->plat->fix_mac_speed))
		priv->plat->fix_mac_speed(priv->plat->bsp_priv,
					  phydev->speed);
}

int gpio_has_init = 0;
static struct gsboard_gpio *bd_gpio_gs = NULL;
static unsigned int gsgpio_count;

#ifdef USE_VIRTUAL_ETH
#ifdef CONFIG_RTL8363_CPUPORT
static void update_link_led(unsigned int port, int onoff)
#else
static void update_link_led(unsigned int port_vid, int onoff)
#endif
{
    unsigned int gpio_pin;
    int led_idx = -1;
	int led_rj45_idx = -1;
    char *led_name;
    char *led_rj45_name;
    int ii;
    
    if (!bd_gpio_gs)
        return;
#ifdef CONFIG_RTL8363_CPUPORT  
    if (port == GXE_LAN_PORT) {
        led_name = LAN_LED_NAME;
		led_rj45_name = LAN_LED_RJ45_NAME;
    } else if (port == GXE_WAN_PORT) {
        led_name = WAN_LED_NAME;
		led_rj45_name = WAN_LED_RJ45_NAME;
    } else 
        return;
#else
    if (port_vid == LAN_VLANID) {
        led_name = LAN_LED_NAME;
		led_rj45_name = LAN_LED_RJ45_NAME;
    } else if (port_vid == WAN_VLANID) {
        led_name = WAN_LED_NAME;
		led_rj45_name = WAN_LED_RJ45_NAME;
    } else 
        return;
#endif    
    for (ii=0; ii<gsgpio_count; ii++) {
        if (!strcmp(led_name, bd_gpio_gs[ii].gpio_name))
            led_idx = ii;
		if (!strcmp(led_rj45_name, bd_gpio_gs[ii].gpio_name))
            led_rj45_idx = ii;
		if ((-1 != led_idx) && (-1 != led_rj45_idx))
			break;
    }

	if ( -1 != led_idx) {
        gpio_pin = (unsigned int)bd_gpio_gs[led_idx].gpio_pin;

        if (bd_gpio_gs[led_idx].skip_init) {
            if (gpio_request(gpio_pin, "gs_board_gpio") < 0) {
                printk("stmmac_set_link_led, request gpio %d failed\n", gpio_pin);
                return;
            }

            if (bd_gpio_gs[led_idx].gpio_dir == 0) { /* input */
                gpio_direction_input(gpio_pin);
            } else {
                gpio_direction_output(gpio_pin, bd_gpio_gs[ii].init_val);
            }

            bd_gpio_gs[led_idx].skip_init = 0;
        }
		gpio_set_value(bd_gpio_gs[led_idx].gpio_pin, onoff);
    }

    if ( -1 != led_rj45_idx) {
        gpio_pin = (unsigned int)bd_gpio_gs[led_rj45_idx].gpio_pin;

        if (bd_gpio_gs[led_rj45_idx].skip_init) {
            if (gpio_request(gpio_pin, "gs_board_gpio") < 0) {
                printk("stmmac_set_link_led, request gpio %d failed\n", gpio_pin);
                return;
            }

            if (bd_gpio_gs[led_rj45_idx].gpio_dir == 0) { /* input */
                gpio_direction_input(gpio_pin);
            } else {
                gpio_direction_output(gpio_pin, bd_gpio_gs[ii].init_val);
            }

            bd_gpio_gs[led_rj45_idx].skip_init = 0;
        }
		gpio_set_value(bd_gpio_gs[led_rj45_idx].gpio_pin, onoff);
    }
}

static void stmmac_set_link_state( struct net_device *dev, unsigned int val)
{
    int ii;
#ifdef CONFIG_RTL8363_CPUPORT
    int port = -1;
#else
    int vlan_id = 0;
#endif    
    if(!dev) 
        return;
    
    for(ii=0;ii<NUM_OF_VDEV;ii++){
       if(virt_ndev[ii].ndev == dev) {
          virt_ndev[ii].link_state = val;
#ifdef CONFIG_RTL8363_CPUPORT
          port = virt_ndev[ii].port;
#else
          vlan_id = virt_ndev[ii].vlan_id;
#endif
          break;
       }
    }
#ifdef CONFIG_RTL8363_CPUPORT 
    if(!gpio_has_init || (port<0))
        return;
    update_link_led(port, val);
#else
    if(!gpio_has_init || !vlan_id)
        return;
    update_link_led(vlan_id, val);
#endif    

}
#endif   //end --#ifdef USE_VIRTUAL_ETH

void gsgpio_init_callback_for_stmmac ( struct gsboard_gpio *bd_gpio, int count )
{
    int ii;
    bd_gpio_gs = bd_gpio;
    gsgpio_count = count;
    gpio_has_init = 1;
#ifdef USE_VIRTUAL_ETH
    /**update link state for LAN and WAN*/
    for(ii=0; ii<NUM_OF_VDEV; ii++){
        #ifdef CONFIG_RTL8363_CPUPORT
        if((virt_ndev[ii].port == GXE_LAN_PORT)||
           (virt_ndev[ii].port == GXE_WAN_PORT))
           update_link_led(virt_ndev[ii].port, virt_ndev[ii].link_state);
        #else
        if((virt_ndev[ii].vlan_id == LAN_VLANID)||
           (virt_ndev[ii].vlan_id == WAN_VLANID))
           update_link_led(virt_ndev[ii].vlan_id, virt_ndev[ii].link_state);
        #endif
    }   
#endif
}
EXPORT_SYMBOL(gsgpio_init_callback_for_stmmac);

#ifdef USE_VIRTUAL_ETH
static void stmmac_link_fixed(struct net_device *dev)
{
	struct stmmac_priv *priv = netdev_priv(dev);
	unsigned long flags;
    u32 ctrl;

	spin_lock_irqsave(&priv->lock, flags);

	ctrl = readl(priv->ioaddr + MAC_CTRL_REG);
    
    /* Flow Control operation */
	priv->hw->mac->flow_ctrl(priv->ioaddr, DUPLEX_FULL,
		FLOW_OFF, PAUSE_TIME);
    
    if (likely(priv->plat->has_gmac))
	   ctrl &= ~priv->hw->link.port;
    if (likely(priv->plat->fix_mac_speed))
       priv->plat->fix_mac_speed(priv->plat->bsp_priv,100);

#ifdef CONFIG_PHY_RTL8363
		/*because half duplex not work*/
    ctrl |= priv->hw->link.duplex;
#endif
	priv->speed = 100;

	writel(ctrl, priv->ioaddr + MAC_CTRL_REG);

	spin_unlock_irqrestore(&priv->lock, flags);
    
	netif_carrier_on(dev);
}
#endif

/**
 * stmmac_adjust_link
 * @dev: net device structure
 * Description: it adjusts the link parameters.
 */
static void stmmac_adjust_link(struct net_device *dev)
{
	struct stmmac_priv *priv = netdev_priv(dev);
	struct phy_device *phydev = priv->phydev;
	unsigned long flags;
	int new_state = 0;
    struct net_device *lan_dev;

#if !defined(USE_VIRTUAL_ETH)    
	unsigned int fc = priv->flow_ctrl, pause_time = priv->pause;
#endif

	if (phydev == NULL)
		return;

	DBG(probe, DEBUG, "stmmac_adjust_link: called.  address %d link %d\n",
	    phydev->addr, phydev->link);

	spin_lock_irqsave(&priv->lock, flags);
	if (phydev->link) {
#if !defined(USE_VIRTUAL_ETH)
		u32 ctrl = readl(priv->ioaddr + MAC_CTRL_REG);

		/* Now we make sure that we can be in full duplex mode.
		 * If not, we operate in half-duplex mode. */
		if (phydev->duplex != priv->oldduplex) {
			new_state = 1;
			if (!(phydev->duplex))
				ctrl &= ~priv->hw->link.duplex;
			else
				ctrl |= priv->hw->link.duplex;
			priv->oldduplex = phydev->duplex;
		}
		/* Flow Control operation */
		if (phydev->pause)
			priv->hw->mac->flow_ctrl(priv->ioaddr, phydev->duplex,
						 fc, pause_time);

		if (phydev->speed != priv->speed) {
			new_state = 1;
			switch (phydev->speed) {
			case 1000:
				if (likely(priv->plat->has_gmac))
					ctrl &= ~priv->hw->link.port;
				stmmac_hw_fix_mac_speed(priv);
				break;
			case 100:
			case 10:
				if (priv->plat->has_gmac) {
					ctrl |= priv->hw->link.port;
					if (phydev->speed == SPEED_100) {
						ctrl |= priv->hw->link.speed;
					} else {
						ctrl &= ~(priv->hw->link.speed);
					}
				} else {
					ctrl &= ~priv->hw->link.port;
				}
				stmmac_hw_fix_mac_speed(priv);
				break;
			default:
				if (netif_msg_link(priv))
					pr_warning("%s: Speed (%d) is not 10"
				       " or 100!\n", dev->name, phydev->speed);
				break;
			}

#ifdef CONFIG_PHY_RTL8363
			/*because half duplex not work*/
			ctrl |= priv->hw->link.duplex;
#endif
			priv->speed = phydev->speed;
		}

		writel(ctrl, priv->ioaddr + MAC_CTRL_REG);
#endif //end ---- #if !defined(USE_VIRTUAL_ETH)
		if (!priv->oldlink) {
			new_state = 1;
			priv->oldlink = 1;
#ifdef USE_VIRTUAL_ETH
            g_link_num++;
            stmmac_set_link_state(dev, 1);
            if(g_route_switch_mode){ /**if switch mode*/
#ifdef CONFIG_RTL8363_CPUPORT
               lan_dev = find_virt_ndev(GXE_LAN_PORT);
#else
               lan_dev = find_virt_ndev(LAN_VLANID);
#endif
               if(lan_dev){
                   if(dev != lan_dev){
                      netif_carrier_on(lan_dev);   
                   }
               }
           }
#endif
		}
	} else if (priv->oldlink) {
		new_state = 1;
		priv->oldlink = 0;
		priv->speed = 0;
		priv->oldduplex = -1;
#ifdef USE_VIRTUAL_ETH
        g_link_num--;
        stmmac_set_link_state(dev, 0);
        if(g_route_switch_mode){ /**if switch mode*/
#ifdef CONFIG_RTL8363_CPUPORT
            lan_dev = find_virt_ndev(GXE_LAN_PORT);
#else
            lan_dev = find_virt_ndev(LAN_VLANID);
#endif
                if(lan_dev){
                    if((dev == lan_dev) && g_link_num){ 
                        netif_carrier_on(lan_dev);   
                    }
                    if (!g_link_num)
                        netif_carrier_off(lan_dev); 
                }
            }
#endif
	}

	if (new_state && netif_msg_link(priv))
		phy_print_status(phydev);

	spin_unlock_irqrestore(&priv->lock, flags);

	DBG(probe, DEBUG, "stmmac_adjust_link: exiting\n");
}

/**
 * stmmac_init_phy - PHY initialization
 * @dev: net device structure
 * Description: it initializes the driver's PHY state, and attaches the PHY
 * to the mac driver.
 *  Return value:
 *  0 on success
 */
static int stmmac_init_phy(struct net_device *dev)
{
	struct stmmac_priv *priv = netdev_priv(dev);
	struct phy_device *phydev;
	char phy_id[MII_BUS_ID_SIZE + 3];
	char bus_id[MII_BUS_ID_SIZE];

	priv->oldlink = 0;
	priv->speed = 0;
	priv->oldduplex = -1;

	if (priv->phy_addr == -1) {
		/* We don't have a PHY, so do nothing */
		return 0;
	}

	snprintf(bus_id, MII_BUS_ID_SIZE, "%x", priv->plat->bus_id);
	snprintf(phy_id, MII_BUS_ID_SIZE + 3, PHY_ID_FMT, bus_id,
		 priv->phy_addr);
	pr_debug("stmmac_init_phy:  trying to attach to %s\n", phy_id);

	phydev = phy_connect(dev, phy_id, &stmmac_adjust_link, 0,
			priv->phy_interface);

	if (IS_ERR(phydev)) {
		pr_err("%s: Could not attach to PHY\n", dev->name);
		return PTR_ERR(phydev);
	}

	/*
	 * Broken HW is sometimes missing the pull-up resistor on the
	 * MDIO line, which results in reads to non-existent devices returning
	 * 0 rather than 0xffff. Catch this here and treat 0 as a non-existent
	 * device as well.
	 * Note: phydev->phy_id is the result of reading the UID PHY registers.
	 */
	if (phydev->phy_id == 0) {
		phy_disconnect(phydev);
		return -ENODEV;
	}
	pr_debug("stmmac_init_phy:  %s: attached to PHY (UID 0x%x)"
	       " Link = %d\n", dev->name, phydev->phy_id, phydev->link);

	priv->phydev = phydev;

	return 0;
}

static inline void stmmac_enable_mac(void __iomem *ioaddr)
{
	u32 value = readl(ioaddr + MAC_CTRL_REG);

	value |= MAC_RNABLE_RX | MAC_ENABLE_TX;
	writel(value, ioaddr + MAC_CTRL_REG);
}

static inline void stmmac_disable_mac(void __iomem *ioaddr)
{
	u32 value = readl(ioaddr + MAC_CTRL_REG);

	value &= ~(MAC_ENABLE_TX | MAC_RNABLE_RX);
	writel(value, ioaddr + MAC_CTRL_REG);
}

/**
 * display_ring
 * @p: pointer to the ring.
 * @size: size of the ring.
 * Description: display all the descriptors within the ring.
 */
static void display_ring(struct dma_desc *p, int size)
{
	struct tmp_s {
		u64 a;
		unsigned int b;
		unsigned int c;
	};
	int i;
	for (i = 0; i < size; i++) {
		struct tmp_s *x = (struct tmp_s *)(p + i);
		pr_info("\t%d [0x%x]: DES0=0x%x DES1=0x%x BUF1=0x%x BUF2=0x%x",
		       i, (unsigned int)virt_to_phys(&p[i]),
		       (unsigned int)(x->a), (unsigned int)((x->a) >> 32),
		       x->b, x->c);
		pr_info("\n");
	}
}

/**
 * init_dma_desc_rings - init the RX/TX descriptor rings
 * @dev: net device structure
 * Description:  this function initializes the DMA RX/TX descriptors
 * and allocates the socket buffers.
 */
static void init_dma_desc_rings(struct net_device *dev)
{
	int i;
	struct stmmac_priv *priv = netdev_priv(dev);
	struct sk_buff *skb;
	unsigned int txsize = priv->dma_tx_size;
	unsigned int rxsize = priv->dma_rx_size;
	unsigned int bfsize = priv->dma_buf_sz;
	int buff2_needed = 0, dis_ic = 0;

	/* Set the Buffer size according to the MTU;
	 * indeed, in case of jumbo we need to bump-up the buffer sizes.
	 */
	if (unlikely(dev->mtu >= BUF_SIZE_8KiB))
		bfsize = BUF_SIZE_16KiB;
	else if (unlikely(dev->mtu >= BUF_SIZE_4KiB))
		bfsize = BUF_SIZE_8KiB;
	else if (unlikely(dev->mtu >= BUF_SIZE_2KiB))
		bfsize = BUF_SIZE_4KiB;
	else if (unlikely(dev->mtu >= DMA_BUFFER_SIZE))
		bfsize = BUF_SIZE_2KiB;
	else
		bfsize = DMA_BUFFER_SIZE;

#ifdef CONFIG_STMMAC_TIMER
	/* Disable interrupts on completion for the reception if timer is on */
	if (likely(priv->tm->enable))
		dis_ic = 1;
#endif
	/* If the MTU exceeds 8k so use the second buffer in the chain */
	if (bfsize >= BUF_SIZE_8KiB)
		buff2_needed = 1;

	DBG(probe, INFO, "stmmac: txsize %d, rxsize %d, bfsize %d\n",
	    txsize, rxsize, bfsize);

	priv->rx_skbuff_dma = kmalloc(rxsize * sizeof(dma_addr_t), GFP_KERNEL);
	priv->rx_skbuff =
	    kmalloc(sizeof(struct sk_buff *) * rxsize, GFP_KERNEL);
	priv->dma_rx =
	    (struct dma_desc *)dma_alloc_coherent(priv->device,
						  rxsize *
						  sizeof(struct dma_desc),
						  &priv->dma_rx_phy,
						  GFP_KERNEL);
	priv->tx_skbuff = kmalloc(sizeof(struct sk_buff *) * txsize,
				       GFP_KERNEL);
	priv->dma_tx =
	    (struct dma_desc *)dma_alloc_coherent(priv->device,
						  txsize *
						  sizeof(struct dma_desc),
						  &priv->dma_tx_phy,
						  GFP_KERNEL);

	if ((priv->dma_rx == NULL) || (priv->dma_tx == NULL)) {
		pr_err("%s:ERROR allocating the DMA Tx/Rx desc\n", __func__);
		return;
	}

	DBG(probe, INFO, "stmmac (%s) DMA desc rings: virt addr (Rx %p, "
	    "Tx %p)\n\tDMA phy addr (Rx 0x%08x, Tx 0x%08x)\n",
	    dev->name, priv->dma_rx, priv->dma_tx,
	    (unsigned int)priv->dma_rx_phy, (unsigned int)priv->dma_tx_phy);

	/* RX INITIALIZATION */
	DBG(probe, INFO, "stmmac: SKB addresses:\n"
			 "skb\t\tskb data\tdma data\n");

	for (i = 0; i < rxsize; i++) {
		struct dma_desc *p = priv->dma_rx + i;

		skb = netdev_alloc_skb_ip_align(dev, bfsize);
		if (unlikely(skb == NULL)) {
			pr_err("%s: Rx init fails; skb is NULL\n", __func__);
			break;
		}
		priv->rx_skbuff[i] = skb;
		priv->rx_skbuff_dma[i] = dma_map_single(priv->device, skb->data,
						bfsize, DMA_FROM_DEVICE);

		p->des2 = priv->rx_skbuff_dma[i];
		if (unlikely(buff2_needed))
			p->des3 = p->des2 + BUF_SIZE_8KiB;
		DBG(probe, INFO, "[%p]\t[%p]\t[%x]\n", priv->rx_skbuff[i],
			priv->rx_skbuff[i]->data, priv->rx_skbuff_dma[i]);
	}
	priv->cur_rx = 0;
	priv->dirty_rx = (unsigned int)(i - rxsize);
	priv->dma_buf_sz = bfsize;
	buf_sz = bfsize;

	/* TX INITIALIZATION */
	for (i = 0; i < txsize; i++) {
		priv->tx_skbuff[i] = NULL;
		priv->dma_tx[i].des2 = 0;
	}
	priv->dirty_tx = 0;
	priv->cur_tx = 0;

	/* Clear the Rx/Tx descriptors */
	priv->hw->desc->init_rx_desc(priv->dma_rx, rxsize, dis_ic);
	priv->hw->desc->init_tx_desc(priv->dma_tx, txsize);

	if (netif_msg_hw(priv)) {
		pr_info("RX descriptor ring:\n");
		display_ring(priv->dma_rx, rxsize);
		pr_info("TX descriptor ring:\n");
		display_ring(priv->dma_tx, txsize);
	}
}

static void dma_free_rx_skbufs(struct stmmac_priv *priv)
{
	int i;

	for (i = 0; i < priv->dma_rx_size; i++) {
		if (priv->rx_skbuff[i]) {
			dma_unmap_single(priv->device, priv->rx_skbuff_dma[i],
					 priv->dma_buf_sz, DMA_FROM_DEVICE);
			dev_kfree_skb_any(priv->rx_skbuff[i]);
		}
		priv->rx_skbuff[i] = NULL;
	}
}

static void dma_free_tx_skbufs(struct stmmac_priv *priv)
{
	int i;

	for (i = 0; i < priv->dma_tx_size; i++) {
		if (priv->tx_skbuff[i] != NULL) {
			struct dma_desc *p = priv->dma_tx + i;
			if (p->des2)
				dma_unmap_single(priv->device, p->des2,
						 priv->hw->desc->get_tx_len(p),
						 DMA_TO_DEVICE);
			dev_kfree_skb_any(priv->tx_skbuff[i]);
			priv->tx_skbuff[i] = NULL;
		}
	}
}

static void free_dma_desc_resources(struct stmmac_priv *priv)
{
	/* Release the DMA TX/RX socket buffers */
	dma_free_rx_skbufs(priv);
	dma_free_tx_skbufs(priv);

	/* Free the region of consistent memory previously allocated for
	 * the DMA */
	dma_free_coherent(priv->device,
			  priv->dma_tx_size * sizeof(struct dma_desc),
			  priv->dma_tx, priv->dma_tx_phy);
	dma_free_coherent(priv->device,
			  priv->dma_rx_size * sizeof(struct dma_desc),
			  priv->dma_rx, priv->dma_rx_phy);
	kfree(priv->rx_skbuff_dma);
	kfree(priv->rx_skbuff);
	kfree(priv->tx_skbuff);
}

/**
 *  stmmac_dma_operation_mode - HW DMA operation mode
 *  @priv : pointer to the private device structure.
 *  Description: it sets the DMA operation mode: tx/rx DMA thresholds
 *  or Store-And-Forward capability.
 */
static void stmmac_dma_operation_mode(struct stmmac_priv *priv)
{
	if (likely((priv->plat->tx_coe) && (!priv->no_csum_insertion))) {
		/* In case of GMAC, SF mode has to be enabled
		 * to perform the TX COE. This depends on:
		 * 1) TX COE if actually supported
		 * 2) There is no bugged Jumbo frame support
		 *    that needs to not insert csum in the TDES.
		 */
		priv->hw->dma->dma_mode(priv->ioaddr,
					SF_DMA_MODE, SF_DMA_MODE);
		tc = SF_DMA_MODE;
	} else
		priv->hw->dma->dma_mode(priv->ioaddr, tc, SF_DMA_MODE);
}

/**
 * stmmac_tx:
 * @priv: private driver structure
 * Description: it reclaims resources after transmission completes.
 */
static void stmmac_tx(struct stmmac_priv *priv)
{
	unsigned int txsize = priv->dma_tx_size;

	while (priv->dirty_tx != priv->cur_tx) {
		int last;
		unsigned int entry = priv->dirty_tx % txsize;
		struct sk_buff *skb = priv->tx_skbuff[entry];
		struct dma_desc *p = priv->dma_tx + entry;

		/* Check if the descriptor is owned by the DMA. */
		if (priv->hw->desc->get_tx_owner(p))
			break;

		/* Verify tx error by looking at the last segment */
		last = priv->hw->desc->get_tx_ls(p);
		if (likely(last)) {
			int tx_error =
				priv->hw->desc->tx_status(&priv->dev->stats,
							  &priv->xstats, p,
							  priv->ioaddr);
			if (likely(tx_error == 0)) {
				priv->dev->stats.tx_packets++;
				priv->xstats.tx_pkt_n++;
			} else
				priv->dev->stats.tx_errors++;
		}
		TX_DBG("%s: curr %d, dirty %d\n", __func__,
			priv->cur_tx, priv->dirty_tx);

		if (likely(p->des2))
			dma_unmap_single(priv->device, p->des2,
					 priv->hw->desc->get_tx_len(p),
					 DMA_TO_DEVICE);
		if (unlikely(p->des3))
			p->des3 = 0;

		if (likely(skb != NULL)) {
			/*
			 * If there's room in the queue (limit it to size)
			 * we add this skb back into the pool,
			 * if it's the right size.
			 */
			if ((skb_queue_len(&priv->rx_recycle) <
				priv->dma_rx_size) &&
				skb_recycle_check(skb, priv->dma_buf_sz))
				__skb_queue_head(&priv->rx_recycle, skb);
			else
				dev_kfree_skb(skb);

			priv->tx_skbuff[entry] = NULL;
		}

		priv->hw->desc->release_tx_desc(p);

		entry = (++priv->dirty_tx) % txsize;
	}
	if (unlikely(netif_queue_stopped(priv->dev) &&
		     stmmac_tx_avail(priv) > STMMAC_TX_THRESH(priv))) {
		netif_tx_lock(priv->dev);
		if (netif_queue_stopped(priv->dev) &&
		     stmmac_tx_avail(priv) > STMMAC_TX_THRESH(priv)) {
			TX_DBG("%s: restart transmit\n", __func__);
			netif_wake_queue(priv->dev);
		}
		netif_tx_unlock(priv->dev);
	}
}

static inline void stmmac_enable_irq(struct stmmac_priv *priv)
{
#ifdef CONFIG_STMMAC_TIMER
	if (likely(priv->tm->enable))
		priv->tm->timer_start(tmrate);
	else
#endif
		priv->hw->dma->enable_dma_irq(priv->ioaddr);
}

static inline void stmmac_disable_irq(struct stmmac_priv *priv)
{
#ifdef CONFIG_STMMAC_TIMER
	if (likely(priv->tm->enable))
		priv->tm->timer_stop();
	else
#endif
		priv->hw->dma->disable_dma_irq(priv->ioaddr);
}

static int stmmac_has_work(struct stmmac_priv *priv)
{
	unsigned int has_work = 0;
	int rxret, tx_work = 0;

	rxret = priv->hw->desc->get_rx_owner(priv->dma_rx +
		(priv->cur_rx % priv->dma_rx_size));

	if (priv->dirty_tx != priv->cur_tx)
		tx_work = 1;

	if (likely(!rxret || tx_work))
		has_work = 1;

	return has_work;
}

static inline void _stmmac_schedule(struct stmmac_priv *priv)
{
	if (likely(stmmac_has_work(priv))) {
		stmmac_disable_irq(priv);
		napi_schedule(&priv->napi);
	}
}

#ifdef CONFIG_STMMAC_TIMER
void stmmac_schedule(struct net_device *dev)
{
	struct stmmac_priv *priv = netdev_priv(dev);

	priv->xstats.sched_timer_n++;

	_stmmac_schedule(priv);
}

static void stmmac_no_timer_started(unsigned int x)
{;
};

static void stmmac_no_timer_stopped(void)
{;
};
#endif

/**
 * stmmac_tx_err:
 * @priv: pointer to the private device structure
 * Description: it cleans the descriptors and restarts the transmission
 * in case of errors.
 */
static void stmmac_tx_err(struct stmmac_priv *priv)
{

	netif_stop_queue(priv->dev);

	priv->hw->dma->stop_tx(priv->ioaddr);
	dma_free_tx_skbufs(priv);
	priv->hw->desc->init_tx_desc(priv->dma_tx, priv->dma_tx_size);
	priv->dirty_tx = 0;
	priv->cur_tx = 0;
	priv->hw->dma->start_tx(priv->ioaddr);

	priv->dev->stats.tx_errors++;
	netif_wake_queue(priv->dev);
}


static void stmmac_dma_interrupt(struct stmmac_priv *priv)
{
	int status;

	status = priv->hw->dma->dma_interrupt(priv->ioaddr, &priv->xstats);
	if (likely(status == handle_tx_rx))
		_stmmac_schedule(priv);

	else if (unlikely(status == tx_hard_error_bump_tc)) {
		/* Try to bump up the dma threshold on this failure */
		if (unlikely(tc != SF_DMA_MODE) && (tc <= 256)) {
			tc += 64;
			priv->hw->dma->dma_mode(priv->ioaddr, tc, SF_DMA_MODE);
			priv->xstats.threshold = tc;
		}
	} else if (unlikely(status == tx_hard_error))
		stmmac_tx_err(priv);
}

/**
 *  stmmac_open - open entry point of the driver
 *  @dev : pointer to the device structure.
 *  Description:
 *  This function is the open entry point of the driver.
 *  Return value:
 *  0 on success and an appropriate (-)ve integer as defined in errno.h
 *  file on failure.
 */
static int stmmac_open(struct net_device *dev)
{
	struct stmmac_priv *priv = netdev_priv(dev);
	int ret;

#ifdef CONFIG_DMW_BOARD_GXE51XX
        extern int gxp22xx_getmac_factory(char *);
        gxp22xx_getmac_factory(dev->dev_addr);
#endif

	/* Check that the MAC address is valid.  If its not, refuse
	 * to bring the device up. The user must specify an
	 * address using the following linux command:
	 *      ifconfig eth0 hw ether xx:xx:xx:xx:xx:xx  */
	if (!is_valid_ether_addr(dev->dev_addr)) {
		random_ether_addr(dev->dev_addr);
		pr_warning("%s: generated random MAC address %pM\n", dev->name,
			dev->dev_addr);
	}

	stmmac_verify_args();

#ifdef CONFIG_STMMAC_TIMER
	priv->tm = kzalloc(sizeof(struct stmmac_timer *), GFP_KERNEL);
	if (unlikely(priv->tm == NULL)) {
		pr_err("%s: ERROR: timer memory alloc failed\n", __func__);
		return -ENOMEM;
	}
	priv->tm->freq = tmrate;

	/* Test if the external timer can be actually used.
	 * In case of failure continue without timer. */
	if (unlikely((stmmac_open_ext_timer(dev, priv->tm)) < 0)) {
		pr_warning("stmmaceth: cannot attach the external timer.\n");
		priv->tm->freq = 0;
		priv->tm->timer_start = stmmac_no_timer_started;
		priv->tm->timer_stop = stmmac_no_timer_stopped;
	} else
		priv->tm->enable = 1;
#endif

#ifndef USE_VIRTUAL_ETH
	ret = stmmac_init_phy(dev);
	if (unlikely(ret)) {
		pr_err("%s: Cannot attach to PHY (error: %d)\n", __func__, ret);
		goto open_error;
	}
#endif

	/* Create and initialize the TX/RX descriptors chains. */
	priv->dma_tx_size = STMMAC_ALIGN(dma_txsize);
	priv->dma_rx_size = STMMAC_ALIGN(dma_rxsize);
	priv->dma_buf_sz = STMMAC_ALIGN(buf_sz);
	init_dma_desc_rings(dev);

	/* DMA initialization and SW reset */
	ret = priv->hw->dma->init(priv->ioaddr, priv->plat->pbl,
				  priv->dma_tx_phy, priv->dma_rx_phy);
	if (ret < 0) {
		pr_err("%s: DMA initialization failed\n", __func__);
		goto open_error;
	}

	/* Copy the MAC addr into the HW  */
	priv->hw->mac->set_umac_addr(priv->ioaddr, dev->dev_addr, 0);
	/* If required, perform hw setup of the bus. */
	if (priv->plat->bus_setup)
		priv->plat->bus_setup(priv->ioaddr);
	/* Initialize the MAC Core */
	priv->hw->mac->core_init(priv->ioaddr);

	priv->rx_coe = priv->hw->mac->rx_coe(priv->ioaddr);
	if (priv->rx_coe)
		pr_info("stmmac: Rx Checksum Offload Engine supported\n");
	if (priv->plat->tx_coe)
		pr_info("\tTX Checksum insertion supported\n");

	/* Initialise the MMC (if present) to disable all interrupts. */
	writel(0xffffffff, priv->ioaddr + MMC_HIGH_INTR_MASK);
	writel(0xffffffff, priv->ioaddr + MMC_LOW_INTR_MASK);

	/* Request the IRQ lines */
	ret = request_irq(dev->irq, stmmac_interrupt,
			 IRQF_SHARED, dev->name, dev);
	if (unlikely(ret < 0)) {
		pr_err("%s: ERROR: allocating the IRQ %d (error: %d)\n",
		       __func__, dev->irq, ret);
		goto open_error;
	}

	/* Enable the MAC Rx/Tx */
	stmmac_enable_mac(priv->ioaddr);

	/* Set the HW DMA mode and the COE */
	stmmac_dma_operation_mode(priv);

	/* Extra statistics */
	memset(&priv->xstats, 0, sizeof(struct stmmac_extra_stats));
	priv->xstats.threshold = tc;

	/* Start the ball rolling... */
	DBG(probe, DEBUG, "%s: DMA RX/TX processes started...\n", dev->name);
	priv->hw->dma->start_tx(priv->ioaddr);
	priv->hw->dma->start_rx(priv->ioaddr);

#ifdef CONFIG_STMMAC_TIMER
	priv->tm->timer_start(tmrate);
#endif
	/* Dump DMA/MAC registers */
	if (netif_msg_hw(priv)) {
		priv->hw->mac->dump_regs(priv->ioaddr);
		priv->hw->dma->dump_regs(priv->ioaddr);
	}
    
#ifndef USE_VIRTUAL_ETH
	if (priv->phydev)
		phy_start(priv->phydev);
#else    
    stmmac_link_fixed(dev);
#endif
	napi_enable(&priv->napi);
	skb_queue_head_init(&priv->rx_recycle);
	netif_start_queue(dev);
   
	return 0;

open_error:
#ifdef CONFIG_STMMAC_TIMER
	kfree(priv->tm);
#endif
#ifndef USE_VIRTUAL_ETH
	if (priv->phydev)
		phy_disconnect(priv->phydev);
#endif
	return ret;
}

/**
 *  stmmac_release - close entry point of the driver
 *  @dev : device pointer.
 *  Description:
 *  This is the stop entry point of the driver.
 */
static int stmmac_release(struct net_device *dev)
{
	struct stmmac_priv *priv = netdev_priv(dev); 
#ifdef USE_VIRTUAL_ETH   
    int ii;
    for(ii=0;ii<NUM_OF_VDEV;ii++){
        if(virt_ndev[ii].ndev){
            if(virt_ndev[ii].ndev->flags & IFF_UP){
              dev_change_flags(virt_ndev[ii].ndev, (virt_ndev[ii].ndev->flags &~(IFF_UP))); 
              netif_stacked_transfer_operstate(dev, virt_ndev[ii].ndev); 
            }
        }
    }
    g_link_num = 0;
#endif

	/* Stop and disconnect the PHY */
	if (priv->phydev) {
		phy_stop(priv->phydev);
		phy_disconnect(priv->phydev);
		priv->phydev = NULL;
	}

	netif_stop_queue(dev);

#ifdef CONFIG_STMMAC_TIMER
	/* Stop and release the timer */
	stmmac_close_ext_timer();
	if (priv->tm != NULL)
		kfree(priv->tm);
#endif
	napi_disable(&priv->napi);
	skb_queue_purge(&priv->rx_recycle);

	/* Free the IRQ lines */
	free_irq(dev->irq, dev);

	/* Stop TX/RX DMA and clear the descriptors */
	priv->hw->dma->stop_tx(priv->ioaddr);
	priv->hw->dma->stop_rx(priv->ioaddr);

	/* Release and free the Rx/Tx resources */
	free_dma_desc_resources(priv);

	/* Disable the MAC Rx/Tx */
	stmmac_disable_mac(priv->ioaddr);

	netif_carrier_off(dev);

	return 0;
}

/*
 * To perform emulated hardware segmentation on skb.
 */
static int stmmac_sw_tso(struct stmmac_priv *priv, struct sk_buff *skb)
{
	struct sk_buff *segs, *curr_skb;
	int gso_segs = skb_shinfo(skb)->gso_segs;

	/* Estimate the number of fragments in the worst case */
	if (unlikely(stmmac_tx_avail(priv) < gso_segs)) {
		netif_stop_queue(priv->dev);
		TX_DBG(KERN_ERR "%s: TSO BUG! Tx Ring full when queue awake\n",
		       __func__);
		if (stmmac_tx_avail(priv) < gso_segs)
			return NETDEV_TX_BUSY;

		netif_wake_queue(priv->dev);
	}
	TX_DBG("\tstmmac_sw_tso: segmenting: skb %p (len %d)\n",
	       skb, skb->len);

	segs = skb_gso_segment(skb, priv->dev->features & ~NETIF_F_TSO);
	if (IS_ERR(segs))
		goto sw_tso_end;

	do {
		curr_skb = segs;
		segs = segs->next;
		TX_DBG("\t\tcurrent skb->len: %d, *curr %p,"
		       "*next %p\n", curr_skb->len, curr_skb, segs);
		curr_skb->next = NULL;
		stmmac_xmit(curr_skb, priv->dev);
	} while (segs);

sw_tso_end:
	dev_kfree_skb(skb);

	return NETDEV_TX_OK;
}

static unsigned int stmmac_handle_jumbo_frames(struct sk_buff *skb,
					       struct net_device *dev,
					       int csum_insertion)
{
	struct stmmac_priv *priv = netdev_priv(dev);
	unsigned int nopaged_len = skb_headlen(skb);
	unsigned int txsize = priv->dma_tx_size;
	unsigned int entry = priv->cur_tx % txsize;
	struct dma_desc *desc = priv->dma_tx + entry;

	if (nopaged_len > BUF_SIZE_8KiB) {

		int buf2_size = nopaged_len - BUF_SIZE_8KiB;

		desc->des2 = dma_map_single(priv->device, skb->data,
					    BUF_SIZE_8KiB, DMA_TO_DEVICE);
		desc->des3 = desc->des2 + BUF_SIZE_4KiB;
		priv->hw->desc->prepare_tx_desc(desc, 1, BUF_SIZE_8KiB,
						csum_insertion);

		entry = (++priv->cur_tx) % txsize;
		desc = priv->dma_tx + entry;

		desc->des2 = dma_map_single(priv->device,
					skb->data + BUF_SIZE_8KiB,
					buf2_size, DMA_TO_DEVICE);
		desc->des3 = desc->des2 + BUF_SIZE_4KiB;
		priv->hw->desc->prepare_tx_desc(desc, 0, buf2_size,
						csum_insertion);
		priv->hw->desc->set_tx_owner(desc);
		priv->tx_skbuff[entry] = NULL;
	} else {
		desc->des2 = dma_map_single(priv->device, skb->data,
					nopaged_len, DMA_TO_DEVICE);
		desc->des3 = desc->des2 + BUF_SIZE_4KiB;
		priv->hw->desc->prepare_tx_desc(desc, 1, nopaged_len,
						csum_insertion);
	}
	return entry;
}

/**
 *  stmmac_xmit:
 *  @skb : the socket buffer
 *  @dev : device pointer
 *  Description : Tx entry point of the driver.
 */
static netdev_tx_t stmmac_xmit(struct sk_buff *skb, struct net_device *dev)
{
	struct stmmac_priv *priv = netdev_priv(dev);
	unsigned int txsize = priv->dma_tx_size;
	unsigned int entry;
	int i, csum_insertion = 0;
	int nfrags = skb_shinfo(skb)->nr_frags;
	struct dma_desc *desc, *first;

	if (unlikely(stmmac_tx_avail(priv) < nfrags + 1)) {
		if (!netif_queue_stopped(dev)) {
			netif_stop_queue(dev);
			/* This is a hard error, log it. */
			pr_err("%s: BUG! Tx Ring full when queue awake\n",
				__func__);
		}
		return NETDEV_TX_BUSY;
	}

	entry = priv->cur_tx % txsize;

#ifdef STMMAC_XMIT_DEBUG
	if ((skb->len > ETH_FRAME_LEN) || nfrags)
		pr_info("stmmac xmit:\n"
		       "\tskb addr %p - len: %d - nopaged_len: %d\n"
		       "\tn_frags: %d - ip_summed: %d - %s gso\n",
		       skb, skb->len, skb_headlen(skb), nfrags, skb->ip_summed,
		       !skb_is_gso(skb) ? "isn't" : "is");
#endif

	if (unlikely(skb_is_gso(skb)))
		return stmmac_sw_tso(priv, skb);

	if (likely((skb->ip_summed == CHECKSUM_PARTIAL))) {
		if (unlikely((!priv->plat->tx_coe) ||
			     (priv->no_csum_insertion)))
			skb_checksum_help(skb);
		else
			csum_insertion = 1;
	}

	desc = priv->dma_tx + entry;
	first = desc;

#ifdef STMMAC_XMIT_DEBUG
	if ((nfrags > 0) || (skb->len > ETH_FRAME_LEN))
		pr_debug("stmmac xmit: skb len: %d, nopaged_len: %d,\n"
		       "\t\tn_frags: %d, ip_summed: %d\n",
		       skb->len, skb_headlen(skb), nfrags, skb->ip_summed);
#endif
	priv->tx_skbuff[entry] = skb;
	if (unlikely(skb->len >= BUF_SIZE_4KiB)) {
		entry = stmmac_handle_jumbo_frames(skb, dev, csum_insertion);
		desc = priv->dma_tx + entry;
	} else {
		unsigned int nopaged_len = skb_headlen(skb);
		desc->des2 = dma_map_single(priv->device, skb->data,
					nopaged_len, DMA_TO_DEVICE);
		priv->hw->desc->prepare_tx_desc(desc, 1, nopaged_len,
						csum_insertion);
	}

	for (i = 0; i < nfrags; i++) {
		skb_frag_t *frag = &skb_shinfo(skb)->frags[i];
		int len = frag->size;

		entry = (++priv->cur_tx) % txsize;
		desc = priv->dma_tx + entry;

		TX_DBG("\t[entry %d] segment len: %d\n", entry, len);
		desc->des2 = dma_map_page(priv->device, frag->page,
					  frag->page_offset,
					  len, DMA_TO_DEVICE);
		priv->tx_skbuff[entry] = NULL;
		priv->hw->desc->prepare_tx_desc(desc, 0, len, csum_insertion);
		priv->hw->desc->set_tx_owner(desc);
	}

	/* Interrupt on completition only for the latest segment */
	priv->hw->desc->close_tx_desc(desc);

#ifdef CONFIG_STMMAC_TIMER
	/* Clean IC while using timer */
	if (likely(priv->tm->enable))
		priv->hw->desc->clear_tx_ic(desc);
#endif
	/* To avoid raise condition */
	priv->hw->desc->set_tx_owner(first);

	priv->cur_tx++;

#ifdef STMMAC_XMIT_DEBUG
	if (netif_msg_pktdata(priv)) {
		pr_info("stmmac xmit: current=%d, dirty=%d, entry=%d, "
		       "first=%p, nfrags=%d\n",
		       (priv->cur_tx % txsize), (priv->dirty_tx % txsize),
		       entry, first, nfrags);
		display_ring(priv->dma_tx, txsize);
		pr_info(">>> frame to be transmitted: ");
		print_pkt(skb->data, skb->len);
	}
#endif

	if (unlikely(stmmac_tx_avail(priv) <= (MAX_SKB_FRAGS + 1))) {
		TX_DBG("%s: stop transmitted packets\n", __func__);
		netif_stop_queue(dev);
	}

	dev->stats.tx_bytes += skb->len;

	priv->hw->dma->enable_dma_transmission(priv->ioaddr);

	return NETDEV_TX_OK;
}

static inline void stmmac_rx_refill(struct stmmac_priv *priv)
{
	unsigned int rxsize = priv->dma_rx_size;
	int bfsize = priv->dma_buf_sz;
	struct dma_desc *p = priv->dma_rx;

	for (; priv->cur_rx - priv->dirty_rx > 0; priv->dirty_rx++) {
		unsigned int entry = priv->dirty_rx % rxsize;
		if (likely(priv->rx_skbuff[entry] == NULL)) {
			struct sk_buff *skb;

			skb = __skb_dequeue(&priv->rx_recycle);
			if (skb == NULL)
				skb = netdev_alloc_skb_ip_align(priv->dev,
								bfsize);

			if (unlikely(skb == NULL))
				break;

			priv->rx_skbuff[entry] = skb;
			priv->rx_skbuff_dma[entry] =
			    dma_map_single(priv->device, skb->data, bfsize,
					   DMA_FROM_DEVICE);

			(p + entry)->des2 = priv->rx_skbuff_dma[entry];
			if (unlikely(priv->plat->has_gmac)) {
				if (bfsize >= BUF_SIZE_8KiB)
					(p + entry)->des3 =
					    (p + entry)->des2 + BUF_SIZE_8KiB;
			}
			RX_DBG(KERN_INFO "\trefill entry #%d\n", entry);
		}
		priv->hw->desc->set_rx_owner(p + entry);
	}
}

static int stmmac_rx(struct stmmac_priv *priv, int limit)
{
	unsigned int rxsize = priv->dma_rx_size;
	unsigned int entry = priv->cur_rx % rxsize;
	unsigned int next_entry;
	unsigned int count = 0;
	struct dma_desc *p = priv->dma_rx + entry;
	struct dma_desc *p_next;
    
#ifdef USE_VIRTUAL_ETH      
    struct net_device *vndev;
    struct stmmac_priv *vpriv;   
    unsigned int use_vndev = 0;
#endif

#ifdef STMMAC_RX_DEBUG
	if (netif_msg_hw(priv)) {
		pr_debug(">>> stmmac_rx: descriptor ring:\n");
		display_ring(priv->dma_rx, rxsize);
	}
#endif
	count = 0;
	while (!priv->hw->desc->get_rx_owner(p)) {
		int status;

		if (count >= limit)
			break;

		count++;

		next_entry = (++priv->cur_rx) % rxsize;
		p_next = priv->dma_rx + next_entry;
		prefetch(p_next);

		/* read the status of the incoming frame */
		status = (priv->hw->desc->rx_status(&priv->dev->stats,
						    &priv->xstats, p));
		if (unlikely(status == discard_frame))
			priv->dev->stats.rx_errors++;
		else {
			struct sk_buff *skb;
			int frame_len;

			frame_len = priv->hw->desc->get_rx_frame_len(p);
			/* ACS is set; GMAC core strips PAD/FCS for IEEE 802.3
			 * Type frames (LLC/LLC-SNAP) */
			if (unlikely(status != llc_snap))
				frame_len -= ETH_FCS_LEN;
#ifdef STMMAC_RX_DEBUG
			if (frame_len > ETH_FRAME_LEN)
				pr_debug("\tRX frame size %d, COE status: %d\n",
					frame_len, status);

			if (netif_msg_hw(priv))
				pr_debug("\tdesc: %p [entry %d] buff=0x%x\n",
					p, entry, p->des2);
#endif
			skb = priv->rx_skbuff[entry];
			if (unlikely(!skb)) {
				pr_err("%s: Inconsistent Rx descriptor chain\n",
					priv->dev->name);
				priv->dev->stats.rx_dropped++;
				break;
			}
			prefetch(skb->data - NET_IP_ALIGN);
			priv->rx_skbuff[entry] = NULL;

			skb_put(skb, frame_len);
			dma_unmap_single(priv->device,
					 priv->rx_skbuff_dma[entry],
					 priv->dma_buf_sz, DMA_FROM_DEVICE);

#ifdef STMMAC_RX_DEBUG
			if (netif_msg_pktdata(priv)) {
				pr_info(" frame received (%dbytes)", frame_len);
				print_pkt(skb->data, frame_len);
			}
#endif
			skb->protocol = eth_type_trans(skb, priv->dev);
            //printk("skb->protocol 0x%x\n", ntohs(skb->protocol));
            // print_pkt(skb->data, frame_len);
#ifdef USE_VIRTUAL_ETH 
#ifdef CONFIG_RTL8363_CPUPORT
            if(skb->protocol == htons(ETH_P_RTL8899) && (skb->len >= sizeof(struct rrcp_hdr))){
                struct rrcp_hdr *rrcphdr;
                u16 port;
                
                rrcphdr = (struct rrcp_hdr *)skb->data;
                port = ntohs(rrcphdr->h_rrcp_portmask);
                
                if(GXE_SVIP_PORT != port && g_route_switch_mode) {
                    port = GXE_LAN_PORT;
                }
                vndev = find_virt_ndev(port);
                if(!vndev){
                   printk_ratelimited("no virtual eth found\n");
                   use_vndev = 0;
                } else {
                   use_vndev = 1;
                   skb->dev = vndev;
                   vpriv = netdev_priv(vndev);
                   memmove(skb->data - 12 + sizeof(struct rrcp_hdr),
                       skb->mac_header, 12);
                   skb_push(skb, 12 - sizeof(struct rrcp_hdr));

                   skb->protocol = eth_type_trans(skb, vpriv->dev);
                   
                   if (!skb) {
                       vpriv->dev->stats.rx_errors++;
                   }
                   skb_checksum_none_assert(skb);
                   netif_receive_skb(skb);    
                   vpriv->dev->stats.rx_packets++;  
                   vpriv->dev->stats.rx_bytes += frame_len;
                }
            } else {
                use_vndev = 0;
            } 
#else
            if(skb->protocol == htons(ETH_P_8021Q) && (skb->len>VLAN_HLEN)){
                __be16 proto;
            	unsigned char *rawp;
                u16 vlan_id;
                u16 vlan_tci;
                struct vlan_hdr *vhdr;
                
                vhdr = (struct vlan_hdr *)skb->data;
                vlan_tci = ntohs(vhdr->h_vlan_TCI);
                vlan_id = vlan_tci & VLAN_VID_MASK;
                vndev = find_virt_ndev(vlan_id);
                if(!vndev){
                   use_vndev = 0;
                } else {
                   use_vndev = 1;
                   skb_pull_rcsum(skb, VLAN_HLEN);
                   skb->dev = vndev;
                   vpriv = netdev_priv(vndev);
            
                	proto = vhdr->h_vlan_encapsulated_proto;
                	if (ntohs(proto) >= 1536) {
                		skb->protocol = proto;
                	} else {
                    	rawp = skb->data;
                    	if (*(unsigned short *)rawp == 0xFFFF)
                    		skb->protocol = htons(ETH_P_802_3);
                    	else
                    		skb->protocol = htons(ETH_P_802_2);
                    }
                    //vlan reorder header
                    if (skb_cow(skb, skb_headroom(skb)) < 0)
                        skb = NULL;
                    if (skb) {
                        /* Lifted from Gleb's VLAN code... */
                        memmove(skb->data - ETH_HLEN,
                            skb->data - VLAN_ETH_HLEN, 12);
                        skb->mac_header += VLAN_HLEN;
                    }
                	if (!skb) {
                		vpriv->dev->stats.rx_errors++;
                	}
                    skb_checksum_none_assert(skb);
            		netif_receive_skb(skb);    
                    vpriv->dev->stats.rx_packets++;  
                    vpriv->dev->stats.rx_bytes += frame_len;
                }
                
            }else {
                use_vndev = 0;
            } 
#endif  //--end #ifdef CONFIG_RTL8363_CPUPORT       
            if(0 == use_vndev){
#endif  //--end #ifdef USE_VIRTUAL_ETH
        		if (unlikely(status == csum_none)) {
        			/* always for the old mac 10/100 */
        			skb_checksum_none_assert(skb);
        			netif_receive_skb(skb);
        		} else {
        			skb->ip_summed = CHECKSUM_UNNECESSARY;
        			napi_gro_receive(&priv->napi, skb);
        		}
#ifdef USE_VIRTUAL_ETH  
            }
#endif
            priv->dev->stats.rx_packets++;
			priv->dev->stats.rx_bytes += frame_len;
		}
		entry = next_entry;
		p = p_next;	/* use prefetched values */
	}

	stmmac_rx_refill(priv);

	priv->xstats.rx_pkt_n += count;

	return count;
}

/**
 *  stmmac_poll - stmmac poll method (NAPI)
 *  @napi : pointer to the napi structure.
 *  @budget : maximum number of packets that the current CPU can receive from
 *	      all interfaces.
 *  Description :
 *   This function implements the the reception process.
 *   Also it runs the TX completion thread
 */
static int stmmac_poll(struct napi_struct *napi, int budget)
{
	struct stmmac_priv *priv = container_of(napi, struct stmmac_priv, napi);
	int work_done = 0;

	priv->xstats.poll_n++;
	stmmac_tx(priv);
	work_done = stmmac_rx(priv, budget);

	if (work_done < budget) {
		napi_complete(napi);
		stmmac_enable_irq(priv);
	}
	return work_done;
}

/**
 *  stmmac_tx_timeout
 *  @dev : Pointer to net device structure
 *  Description: this function is called when a packet transmission fails to
 *   complete within a reasonable tmrate. The driver will mark the error in the
 *   netdev structure and arrange for the device to be reset to a sane state
 *   in order to transmit a new packet.
 */
static void stmmac_tx_timeout(struct net_device *dev)
{
	struct stmmac_priv *priv = netdev_priv(dev);

	/* Clear Tx resources and restart transmitting again */
	stmmac_tx_err(priv);
}

/* Configuration changes (passed on by ifconfig) */
static int stmmac_config(struct net_device *dev, struct ifmap *map)
{
	if (dev->flags & IFF_UP)	/* can't act on a running interface */
		return -EBUSY;

	/* Don't allow changing the I/O address */
	if (map->base_addr != dev->base_addr) {
		pr_warning("%s: can't change I/O address\n", dev->name);
		return -EOPNOTSUPP;
	}

	/* Don't allow changing the IRQ */
	if (map->irq != dev->irq) {
		pr_warning("%s: can't change IRQ number %d\n",
		       dev->name, dev->irq);
		return -EOPNOTSUPP;
	}

	/* ignore other fields */
	return 0;
}

/**
 *  stmmac_multicast_list - entry point for multicast addressing
 *  @dev : pointer to the device structure
 *  Description:
 *  This function is a driver entry point which gets called by the kernel
 *  whenever multicast addresses must be enabled/disabled.
 *  Return value:
 *  void.
 */
static void stmmac_multicast_list(struct net_device *dev)
{
	struct stmmac_priv *priv = netdev_priv(dev);

	spin_lock(&priv->lock);
	priv->hw->mac->set_filter(dev);
	spin_unlock(&priv->lock);
}

/**
 *  stmmac_change_mtu - entry point to change MTU size for the device.
 *  @dev : device pointer.
 *  @new_mtu : the new MTU size for the device.
 *  Description: the Maximum Transfer Unit (MTU) is used by the network layer
 *  to drive packet transmission. Ethernet has an MTU of 1500 octets
 *  (ETH_DATA_LEN). This value can be changed with ifconfig.
 *  Return value:
 *  0 on success and an appropriate (-)ve integer as defined in errno.h
 *  file on failure.
 */
static int stmmac_change_mtu(struct net_device *dev, int new_mtu)
{
	struct stmmac_priv *priv = netdev_priv(dev);
	int max_mtu;

	if (netif_running(dev)) {
		pr_err("%s: must be stopped to change its MTU\n", dev->name);
		return -EBUSY;
	}

	if (priv->plat->has_gmac)
		max_mtu = JUMBO_LEN;
	else
		max_mtu = ETH_DATA_LEN;

	if ((new_mtu < 46) || (new_mtu > max_mtu)) {
		pr_err("%s: invalid MTU, max MTU is: %d\n", dev->name, max_mtu);
		return -EINVAL;
	}

	/* Some GMAC devices have a bugged Jumbo frame support that
	 * needs to have the Tx COE disabled for oversized frames
	 * (due to limited buffer sizes). In this case we disable
	 * the TX csum insertionin the TDES and not use SF. */
	if ((priv->plat->bugged_jumbo) && (priv->dev->mtu > ETH_DATA_LEN))
		priv->no_csum_insertion = 1;
	else
		priv->no_csum_insertion = 0;

	dev->mtu = new_mtu;

	return 0;
}

static irqreturn_t stmmac_interrupt(int irq, void *dev_id)
{
	struct net_device *dev = (struct net_device *)dev_id;
	struct stmmac_priv *priv = netdev_priv(dev);

	if (unlikely(!dev)) {
		pr_err("%s: invalid dev pointer\n", __func__);
		return IRQ_NONE;
	}

	if (priv->plat->has_gmac)
		/* To handle GMAC own interrupts */
		priv->hw->mac->host_irq_status((void __iomem *) dev->base_addr);

	stmmac_dma_interrupt(priv);

	return IRQ_HANDLED;
}

#ifdef CONFIG_NET_POLL_CONTROLLER
/* Polling receive - used by NETCONSOLE and other diagnostic tools
 * to allow network I/O with interrupts disabled. */
static void stmmac_poll_controller(struct net_device *dev)
{
	disable_irq(dev->irq);
	stmmac_interrupt(dev->irq, dev);
	enable_irq(dev->irq);
}
#endif

/**
 *  stmmac_ioctl - Entry point for the Ioctl
 *  @dev: Device pointer.
 *  @rq: An IOCTL specefic structure, that can contain a pointer to
 *  a proprietary structure used to pass information to the driver.
 *  @cmd: IOCTL command
 *  Description:
 *  Currently there are no special functionality supported in IOCTL, just the
 *  phy_mii_ioctl(...) can be invoked.
 */
static int stmmac_ioctl(struct net_device *dev, struct ifreq *rq, int cmd)
{
	struct stmmac_priv *priv = netdev_priv(dev);
	int ret;

	if (!netif_running(dev))
		return -EINVAL;

	if (!priv->phydev)
		return -EINVAL;

	spin_lock(&priv->lock);
	ret = phy_mii_ioctl(priv->phydev, rq, cmd);
	spin_unlock(&priv->lock);

	return ret;
}

#ifdef STMMAC_VLAN_TAG_USED
static void stmmac_vlan_rx_register(struct net_device *dev,
				    struct vlan_group *grp)
{
	struct stmmac_priv *priv = netdev_priv(dev);

	DBG(probe, INFO, "%s: Setting vlgrp to %p\n", dev->name, grp);

	spin_lock(&priv->lock);
	priv->vlgrp = grp;
	spin_unlock(&priv->lock);
}
#endif

#ifdef USE_VIRTUAL_ETH 

static int v_stmmac_open(struct net_device *dev)
{
    int ii;
    int ret;
    struct stmmac_priv *priv = netdev_priv(dev);
#ifdef CONFIG_DMW_BOARD_GXE51XX
    extern int gxp22xx_getmac_factory(char *);
#endif
    
    if(!(real_ndev->flags & IFF_UP)){
        dev_change_flags(real_ndev, real_ndev->flags|IFF_UP);  
        netif_stacked_transfer_operstate(dev, real_ndev);
    }

#ifdef CONFIG_DMW_BOARD_GXE51XX
    gxp22xx_getmac_factory(dev->dev_addr);
#endif

	/* Check that the MAC address is valid.  If its not, refuse
	 * to bring the device up. The user must specify an
	 * address using the following linux command:
	 *      ifconfig eth0 hw ether xx:xx:xx:xx:xx:xx  */
	if (!is_valid_ether_addr(dev->dev_addr)) {
		random_ether_addr(dev->dev_addr);
		pr_warning("%s: generated random MAC address %pM\n", dev->name,
			dev->dev_addr);
	}
#ifdef CONFIG_RTL8363_CPUPORT
    ret = find_vndev_port(dev);
    if(ret != GXE_SVIP_PORT)
#else
    ret = find_vndev_vid(dev);
    if(ret != SVIP_VLANID)
#endif
    {
	    ret = stmmac_init_phy(dev);
        if (unlikely(ret)) {
        	pr_err("%s: Cannot attach to PHY (error: %d)\n", __func__, ret);
            return ret;
        }
    } else {
        netif_carrier_on(dev);
    }

	/* Extra statistics */
	memset(&priv->xstats, 0, sizeof(struct stmmac_extra_stats));
	priv->xstats.threshold = tc;
    
	if (priv->phydev)
		phy_start(priv->phydev);

	netif_start_queue(dev);

    for(ii=0;ii<NUM_OF_VDEV;ii++){
       if(virt_ndev[ii].ndev == dev)
           virt_ndev[ii].opened = 1;
    }
    
	return 0;
    
}

static int v_stmmac_release(struct net_device *dev)
{
    struct stmmac_priv *priv = netdev_priv(dev);
    int ii;
    
	/* Stop and disconnect the PHY */
	if (priv->phydev) {
		phy_stop(priv->phydev);
		phy_disconnect(priv->phydev);
		priv->phydev = NULL;
	}

	netif_stop_queue(dev);

	netif_carrier_off(dev);
    
    for(ii=0;ii<NUM_OF_VDEV;ii++){
       if(virt_ndev[ii].ndev == dev)
           virt_ndev[ii].opened = 0;
    }
    if(priv->oldlink){
        g_link_num--;
        if(g_link_num < 0)g_link_num = 0;
        priv->oldlink = 0;
    }
	return 0;
}
static netdev_tx_t v_stmmac_xmit(struct sk_buff *skb, struct net_device *dev)
{
	unsigned int len;
	int ret;
#ifdef CONFIG_RTL8363_CPUPORT 
	u16 port, portmask = 0;
    __be16 *proto;
    struct rrcp_hdr *rrcphdr;
    
    int cptag_len = sizeof(struct rrcp_hdr) + 2;
    
    port = find_vndev_port(dev);

    if (skb_cow(skb, skb_headroom(skb) + cptag_len)) {
       if (net_ratelimit())
			pr_err("v_stmmac_xmit: out of memory\n");
		kfree_skb(skb);
        
		return NETDEV_TX_OK;
    }

    memmove(skb->data - cptag_len, skb->data, 12);
    skb_push(skb, cptag_len);
    
    proto = ( __be16 *)(skb->data + 12);
    rrcphdr = (struct rrcp_hdr *)(skb->data + ETH_HLEN);
    
    *proto = htons(ETH_P_RTL8899);
    rrcphdr->h_rrcp_fid_prior = htons(0x0);
    rrcphdr->h_rrcp_proto_reason = htons(0x0400);
    
    if(GXE_SVIP_PORT != port && g_route_switch_mode) {
        /**for switch mode forward to LAN and WAN ports*/
        portmask = ((1 << GXE_LAN_PORT) | (1 << GXE_WAN_PORT));
    } else {
        portmask = (1 << port);
    }
    rrcphdr->h_rrcp_portmask = htons(portmask);
    skb_reset_mac_header(skb);
#else
	u16 vlan_tci;

    vlan_tci = find_vndev_vid(dev);
    if(likely(vlan_tci>0)){
        skb->vlan_tci = VLAN_TAG_PRESENT | vlan_tci; 
    } else {
        if (net_ratelimit())
            pr_warning("cannot find VID of virtual ethernet device \n");
    }
#endif
	skb_set_dev(skb, real_ndev);
    
	len = skb->len;
	ret = dev_queue_xmit(skb);

	if (likely(ret == NET_XMIT_SUCCESS || ret == NET_XMIT_CN)) {
		dev->stats.tx_packets++;
		dev->stats.tx_bytes += len;
	} else {
		dev->stats.tx_dropped++;
	}

	return ret;
}

/**
 * v_stmmac_change_mtu - set new MTU size
 * @dev: network device
 * @new_mtu: new Maximum Transfer Unit
 *
 * Allow changing MTU size.
 */
static int v_stmmac_change_mtu(struct net_device *dev, int new_mtu)
{
	if (new_mtu < 68 || new_mtu > ETH_DATA_LEN)
		return -EINVAL;
	dev->mtu = new_mtu;
	return 0;
}

static void v_stmmac_multicast_list(struct net_device *dev)
{
    struct stmmac_priv *rdev_priv;

    if(real_ndev){ 
        rdev_priv = netdev_priv(real_ndev);
    	spin_lock(&rdev_priv->lock);
    	rdev_priv->hw->mac->set_filter(real_ndev);
    	spin_unlock(&rdev_priv->lock);
    }
}

static void v_stmmac_tx_timeout(struct net_device *dev)
{
}

static int v_stmmac_ioctl(struct net_device *dev, struct ifreq *rq, int cmd)
{
    return 0;
}

#ifdef STMMAC_VLAN_TAG_USED
static void v_stmmac_vlan_rx_register(struct net_device *dev,
				    struct vlan_group *grp)
{
	struct stmmac_priv *priv = netdev_priv(dev);

	DBG(probe, INFO, "%s: Setting vlgrp to %p\n", dev->name, grp);

	spin_lock(&priv->lock);
	priv->vlgrp = grp;
	spin_unlock(&priv->lock);
}
#endif
/* Configuration changes (passed on by ifconfig) */
static int v_stmmac_config(struct net_device *dev, struct ifmap *map)
{
	if (dev->flags & IFF_UP)	/* can't act on a running interface */
		return -EBUSY;

	/* Don't allow changing the I/O address */
	if (map->base_addr != dev->base_addr) {
		pr_warning("%s: can't change I/O address\n", dev->name);
		return -EOPNOTSUPP;
	}

	/* Don't allow changing the IRQ */
	if (map->irq != dev->irq) {
		pr_warning("%s: can't change IRQ number %d\n",
		       dev->name, dev->irq);
		return -EOPNOTSUPP;
	}

	/* ignore other fields */
	return 0;
}

static const struct net_device_ops v_stmmac_netdev_ops = {
    .ndo_open = v_stmmac_open,
    .ndo_start_xmit = v_stmmac_xmit,
    .ndo_stop = v_stmmac_release,
    .ndo_change_mtu = v_stmmac_change_mtu,
    .ndo_set_multicast_list = v_stmmac_multicast_list,
    .ndo_tx_timeout = v_stmmac_tx_timeout,
    .ndo_do_ioctl = v_stmmac_ioctl,
    .ndo_set_config = v_stmmac_config,
#ifdef STMMAC_VLAN_TAG_USED
    .ndo_vlan_rx_register = v_stmmac_vlan_rx_register,
#endif
    .ndo_set_mac_address = eth_mac_addr,
  
};

#endif 

static const struct net_device_ops stmmac_netdev_ops = {
	.ndo_open = stmmac_open,
	.ndo_start_xmit = stmmac_xmit,
	.ndo_stop = stmmac_release,
	.ndo_change_mtu = stmmac_change_mtu,
	.ndo_set_multicast_list = stmmac_multicast_list,
	.ndo_tx_timeout = stmmac_tx_timeout,
	.ndo_do_ioctl = stmmac_ioctl,
	.ndo_set_config = stmmac_config,
#ifdef STMMAC_VLAN_TAG_USED
	.ndo_vlan_rx_register = stmmac_vlan_rx_register,
#endif
#ifdef CONFIG_NET_POLL_CONTROLLER
	.ndo_poll_controller = stmmac_poll_controller,
#endif
	.ndo_set_mac_address = eth_mac_addr,
};

#ifdef USE_VIRTUAL_ETH
extern int rtk_port_isolation_set(struct mii_bus *bus, unsigned int port, unsigned int portmask);
extern int rtk_port_isolation_get(struct mii_bus *bus, unsigned int port, unsigned int *pPortmask);

extern int gxe_board_vlan_set(struct mii_bus *bus, int lan_vid, int wlan_vid, int svip_vid);
extern int gxe_board_cpuport_set(struct mii_bus *bus, int smode);
extern int gxe_board_cpuport_get(struct mii_bus *bus);

static int gxe_route_switch_mode_proc_read (char *page, char **start,
    off_t off, int count, int *eof, void *data)
{
    char *p = page;
    int  len = 0;

    p += sprintf(p, "%x\n", g_route_switch_mode? 1 : 0); 

    len = (p - page);
    if (len <= off+count) *eof = 1;
    *start = page + off;
    len -= off;
    if (len>count) len = count;
    if (len<0) len = 0;

    return (len);
}

static int gxe_route_switch_mode_proc_write (struct file *file, 
    const char *buffer, unsigned long count, void *data)
{
    char state_string[12] = {'\0'};

    if ((count > sizeof(state_string) - 1))
        return (-EINVAL);

    if (copy_from_user(state_string, buffer, count))
        return (-EFAULT);

    state_string[count] = '\0';
    
    if(!g_mdio_priv.mii){
        printk("rtl8363 mdio not be registered!!can not set route swith mode!\n");
        return (count);
    }
    
#if CONFIG_RTL8363_CPUPORT
    if (state_string[0] == '1') {
        g_route_switch_mode = 1;/**swith mode*/
        if(rtk_port_isolation_set(g_mdio_priv.mii, 1, 0x24))
             printk("PORT1(LAN) isolation Error(portmask 0x2f)!!\n");
        if(rtk_port_isolation_set(g_mdio_priv.mii, 2, 0x22))
             printk("PORT2(WAN) isolation Error(portmask 0x2f)!!\n");
    } else {
        g_route_switch_mode = 0;/**route mode*/
        if(rtk_port_isolation_set(g_mdio_priv.mii, 1, 0x20))
             printk("PORT1(LAN) isolation Error(portmask 0x20)!!\n");
        if(rtk_port_isolation_set(g_mdio_priv.mii, 2, 0x20))
             printk("PORT2(WAN) isolation Error(portmask 0x20)!!\n");
    }
#else
    if (state_string[0] == '1') {
        g_route_switch_mode = 1;/**swith mode, WAN_VLANID=LAN_VLANID*/
        gxe_board_vlan_set(g_mdio_priv.mii, LAN_VLANID, LAN_VLANID, SVIP_VLANID);
    } else {
        g_route_switch_mode = 0;/**route mode*/
        gxe_board_vlan_set(g_mdio_priv.mii, LAN_VLANID, WAN_VLANID, SVIP_VLANID);
    }
#endif

    return (count);
}

static int gxe_port_isolation_proc_read (char *page, char **start,
    off_t off, int count, int *eof, void *data)
{
    int  len = 0;
    int portmask = 0;

    if(!g_mdio_priv.mii){
        printk(KERN_ERR "rtl8363 mdio not be registered!!can not get port isolation!\n");
        *eof = 1;
        return 0;
    }
    
    rtk_port_isolation_get(g_mdio_priv.mii, GXE_LAN_PORT, &portmask);
    len += snprintf(page+len, PAGE_SIZE - len, "[LAN]port %d Portmask 0x%x\n", 
        GXE_LAN_PORT, portmask);
    
    rtk_port_isolation_get(g_mdio_priv.mii, GXE_WAN_PORT, &portmask);
    len += snprintf(page+len, PAGE_SIZE - len, "[WAN]port %d Portmask 0x%x\n", 
        GXE_WAN_PORT, portmask);

    rtk_port_isolation_get(g_mdio_priv.mii, GXE_DMW96_PORT, &portmask);
    len += snprintf(page+len, PAGE_SIZE - len, "[DMW96]port %d Portmask 0x%x\n", 
        GXE_DMW96_PORT, portmask);
    
    rtk_port_isolation_get(g_mdio_priv.mii, GXE_SVIP_PORT, &portmask);
    len += snprintf(page+len, PAGE_SIZE - len, "[SVIP]port %d Portmask 0x%x\n", 
        GXE_SVIP_PORT, portmask);

    if (len <= off+count) 
        *eof = 1;
    *start = page + off;
    len -= off;
    
    if (len>count)
        len = count;
    if (len<0) 
        len = 0;

    return (len);
}

static int gxe_port_isolation_proc_write (struct file *file, 
    const char *buffer, unsigned long count, void *data)
{
    char iso_string[1024] = {'\0'};
    char *tok, *tok2, *endptr;
    int res;
    int port, portmask;
    char *s;

    if ((count > sizeof(iso_string) - 1))
        return (-EINVAL);
    
    res = copy_from_user(iso_string, buffer, count);
    if (res)
        return (-EFAULT);

    s = (char *)&iso_string[0];

    tok = strsep(&s, " ");
    if (!tok)
        return (-EINVAL);

    port = simple_strtoul((const char *)tok, &endptr, 16);
    
    tok2 = strsep(&s, " ");
    if (!tok2)
        return (-EINVAL);

    portmask = simple_strtoul((const char *)tok2, &endptr, 16);

    if(!g_mdio_priv.mii){
        printk("rtl8363 mdio not be registered!!can not set port isolation!\n");
        return (count);
    }
    
    rtk_port_isolation_set(g_mdio_priv.mii, port, portmask);

    return (count);
}

static int gxe_cpuport_mode_proc_read (char *page, char **start,
    off_t off, int count, int *eof, void *data)
{
    char *p = page;
    int  len = 0;
    int smode = -1;
    
    if(!g_mdio_priv.mii){
        printk("rtl8363 mdio not be registered!!\n");
        return (count);
    }

    smode = gxe_board_cpuport_get(g_mdio_priv.mii);
    if(smode<0){
        smode = 0;
    }
    p += sprintf(p, "%d\n", smode & 0xff); 

    len = (p - page);
    if (len <= off+count) *eof = 1;
    *start = page + off;
    len -= off;
    if (len>count) len = count;
    if (len<0) len = 0;

    return (len);
}

static int gxe_cpuport_mode_proc_write (struct file *file, 
    const char *buffer, unsigned long count, void *data)
{
    char *endptr;
    int smode;
    char smode_string[12] = {'\0'};
    
    if ((count > sizeof(smode_string) - 1))
        return (-EINVAL);

    if (copy_from_user(smode_string, buffer, count))
        return (-EFAULT);

    smode_string[count] = '\0';

    if(!g_mdio_priv.mii){
        printk("rtl8363 mdio not be registered!!\n");
        return (count);
    }
    
    smode = simple_strtoul((const char *)smode_string, &endptr, 10);
    
    gxe_board_cpuport_set(g_mdio_priv.mii, smode);

    return (count);
}


struct proc_dir_entry *rtl8363_router_dir;
static int gxe_rtl8363_install_proc_entry (void)
{
    struct proc_dir_entry   *ent;
    int ret; 

    rtl8363_router_dir = proc_mkdir( "rtl8363_router", NULL);
    
    if(rtl8363_router_dir == NULL) {
        ret = -ENOMEM;
        goto out; 
    }    

    ent = create_proc_entry("route_switch_mode", S_IFREG|S_IRUGO, rtl8363_router_dir);
    if (!ent) {
          printk("create proc entry <route_switch_mode> failed\n");
          ret = -1;
          goto create_proc_fail;
    }    
    ent->read_proc  = gxe_route_switch_mode_proc_read;
    ent->write_proc = gxe_route_switch_mode_proc_write;

    ent = create_proc_entry("port_isolation", S_IFREG|S_IRUGO, rtl8363_router_dir);
    if (!ent) {
          printk("create proc entry <port_isolation> failed\n");
          ret = -1;
          goto create_proc_fail1;
    }    
    ent->read_proc  = gxe_port_isolation_proc_read;
    ent->write_proc = gxe_port_isolation_proc_write;

    ent = create_proc_entry("cpu_port_mode", S_IFREG|S_IRUGO, rtl8363_router_dir);
    if (!ent) {
          printk("create proc entry <cpu_port_mode> failed\n");
          ret = -1;
          goto create_proc_fail2;
    }    
    ent->read_proc  = gxe_cpuport_mode_proc_read;
    ent->write_proc = gxe_cpuport_mode_proc_write;

    return 0;
create_proc_fail2:
    remove_proc_entry("port_isolation", rtl8363_router_dir);
create_proc_fail1:
    remove_proc_entry("route_switch_mode", rtl8363_router_dir);
create_proc_fail:     
    remove_proc_entry("rtl8363_router", NULL);
    rtl8363_router_dir = NULL;
out:
    return ret; 
}

static void gxe_rtl8363_remove_proc_entry (void)
{
    if(rtl8363_router_dir){
        remove_proc_entry("cpu_port_mode", rtl8363_router_dir);
        remove_proc_entry("port_isolation", rtl8363_router_dir);
        remove_proc_entry("route_switch_mode", rtl8363_router_dir);
        remove_proc_entry("rtl8363_router", NULL);
        rtl8363_router_dir = NULL;
   }
}

#endif

/**
 * stmmac_probe - Initialization of the adapter .
 * @dev : device pointer
 * Description: The function initializes the network device structure for
 * the STMMAC driver. It also calls the low level routines
 * in order to init the HW (i.e. the DMA engine)
 */
static int stmmac_probe(struct net_device *dev)
{
	int ret = 0;
	struct stmmac_priv *priv = netdev_priv(dev);

	ether_setup(dev);

	dev->netdev_ops = &stmmac_netdev_ops;
#ifndef USE_VIRTUAL_ETH      
	stmmac_set_ethtool_ops(dev);
#endif
	dev->features |= NETIF_F_SG | NETIF_F_HIGHDMA |
		NETIF_F_IP_CSUM | NETIF_F_IPV6_CSUM;
	dev->watchdog_timeo = msecs_to_jiffies(watchdog);
#ifdef STMMAC_VLAN_TAG_USED
	/* Both mac100 and gmac support receive VLAN tag detection */
	dev->features |= NETIF_F_HW_VLAN_RX;
#endif
	priv->msg_enable = netif_msg_init(debug, default_msg_level);

	if (flow_ctrl)
		priv->flow_ctrl = FLOW_AUTO;	/* RX/TX pause on */

	priv->pause = pause;
	netif_napi_add(dev, &priv->napi, stmmac_poll, 64);

	/* Get the MAC address */
	priv->hw->mac->get_umac_addr((void __iomem *) dev->base_addr,
				     dev->dev_addr, 0);

	if (!is_valid_ether_addr(dev->dev_addr))
		pr_warning("\tno valid MAC address;"
			"please, use ifconfig or nwhwconfig!\n");

	spin_lock_init(&priv->lock);

	ret = register_netdev(dev);
	if (ret) {
		pr_err("%s: ERROR %i registering the device\n",
		       __func__, ret);
		return -ENODEV;
	}

	DBG(probe, DEBUG, "%s: Scatter/Gather: %s - HW checksums: %s\n",
	    dev->name, (dev->features & NETIF_F_SG) ? "on" : "off",
	    (dev->features & NETIF_F_IP_CSUM) ? "on" : "off");

	return ret;
}

/**
 * stmmac_mac_device_setup
 * @dev : device pointer
 * Description: select and initialise the mac device (mac100 or Gmac).
 */
static int stmmac_mac_device_setup(struct net_device *dev)
{
	struct stmmac_priv *priv = netdev_priv(dev);

	struct mac_device_info *device;

	if (priv->plat->has_gmac)
		device = dwmac1000_setup(priv->ioaddr);
	else
		device = dwmac100_setup(priv->ioaddr);

	if (!device)
		return -ENOMEM;

	if (priv->plat->enh_desc) {
		device->desc = &enh_desc_ops;
		pr_info("\tEnhanced descriptor structure\n");
	} else
		device->desc = &ndesc_ops;

	priv->hw = device;

	if (device_can_wakeup(priv->device)) {
		priv->wolopts = WAKE_MAGIC; /* Magic Frame as default */
		enable_irq_wake(dev->irq);
	}

	return 0;
}

static int stmmacphy_dvr_probe(struct platform_device *pdev)
{
	struct plat_stmmacphy_data *plat_dat = pdev->dev.platform_data;

	pr_debug("stmmacphy_dvr_probe: added phy for bus %d\n",
	       plat_dat->bus_id);

	return 0;
}

static int stmmacphy_dvr_remove(struct platform_device *pdev)
{
	return 0;
}

static struct platform_driver stmmacphy_driver = {
	.driver = {
		   .name = PHY_RESOURCE_NAME,
		   },
	.probe = stmmacphy_dvr_probe,
	.remove = stmmacphy_dvr_remove,
};

#ifndef USE_VIRTUAL_ETH
/**
 * stmmac_associate_phy
 * @dev: pointer to device structure
 * @data: points to the private structure.
 * Description: Scans through all the PHYs we have registered and checks if
 * any are associated with our MAC.  If so, then just fill in
 * the blanks in our local context structure
 */
static int stmmac_associate_phy(struct device *dev, void *data)
{
	struct stmmac_priv *priv = (struct stmmac_priv *)data;
	struct plat_stmmacphy_data *plat_dat = dev->platform_data;

	DBG(probe, DEBUG, "%s: checking phy for bus %d\n", __func__,
		plat_dat->bus_id);

	/* Check that this phy is for the MAC being initialised */
	if (priv->plat->bus_id != plat_dat->bus_id)
		return 0;

	/* OK, this PHY is connected to the MAC.
	   Go ahead and get the parameters */
	DBG(probe, DEBUG, "%s: OK. Found PHY config\n", __func__);
	priv->phy_irq =
	    platform_get_irq_byname(to_platform_device(dev), "phyirq");
	DBG(probe, DEBUG, "%s: PHY irq on bus %d is %d\n", __func__,
	    plat_dat->bus_id, priv->phy_irq);

	/* Override with kernel parameters if supplied XXX CRS XXX
	 * this needs to have multiple instances */
	if ((phyaddr >= 0) && (phyaddr <= 31))
		plat_dat->phy_addr = phyaddr;

	priv->phy_addr = plat_dat->phy_addr;
	priv->phy_mask = plat_dat->phy_mask;
	priv->phy_interface = plat_dat->interface;
	priv->phy_reset = plat_dat->phy_reset;

	DBG(probe, DEBUG, "%s: exiting\n", __func__);
	return 1;	/* forces exit of driver_for_each_device() */
}
#endif

#ifdef USE_VIRTUAL_ETH
static int stmmac_veth_associate_phy(struct device *dev, void *data)
{
    struct stmmac_mdio_priv *mdio_priv = (struct stmmac_mdio_priv *)data;
	struct plat_stmmacphy_data *plat_dat = dev->platform_data;
    int phy_idx = mdio_priv->has_nphys;
    
	DBG(probe, DEBUG, "%s: checking phy for bus %d\n", __func__,
		plat_dat->bus_id);

    if(mdio_priv->has_nphys >= GXE_MAX_PHY_NUM){
        return 1; /* forces exit of driver_for_each_device() */
    }

	DBG(probe, DEBUG, "%s: OK. Found PHY config\n", __func__);
	mdio_priv->phy_irq[phy_idx] =
	    platform_get_irq_byname(to_platform_device(dev), "phyirq");
	DBG(probe, DEBUG, "%s: PHY irq on bus %d is %d\n", __func__,
	    plat_dat->bus_id, mdio_priv->phy_irq[phy_idx]);

	mdio_priv->phy_addr[phy_idx] = plat_dat->phy_addr;
	mdio_priv->phy_interface[phy_idx] = plat_dat->interface;

    /*use same reset and phy_mask */
	mdio_priv->phy_reset = plat_dat->phy_reset;
    mdio_priv->phy_mask = plat_dat->phy_mask;
    mdio_priv->has_nphys++;
	return 0; 
}

static int add_virt_ndev(struct platform_device *pdev)
{
    int ii;
    int ret = 0;
    
    struct net_device *vndev[3];
    struct stmmac_priv *vpriv[3];
    struct stmmac_priv *rdev_priv;
    struct plat_stmmacenet_data *plat_dat = pdev->dev.platform_data;

    for(ii=0;ii<NUM_OF_VDEV;ii++){
#ifdef CONFIG_RTL8363_CPUPORT
        if(virt_ndev[ii].port == GXE_SVIP_PORT)
#else
        if(virt_ndev[ii].vlan_id == SVIP_VLANID)
#endif
            vndev[ii] = alloc_netdev_mqs(sizeof(struct stmmac_priv), "svip%d", ether_setup, 1, 1);
        else
            vndev[ii] = alloc_etherdev(sizeof(struct stmmac_priv));
        if (!vndev[ii]) {
            pr_err("%s: ERROR: allocating the device\n", __func__);
            return -ENOMEM;
        }
        virt_ndev[ii].ndev = vndev[ii];
        SET_NETDEV_DEV(vndev[ii], &pdev->dev);
        vpriv[ii] = netdev_priv(vndev[ii]);
	    vpriv[ii]->dev = vndev[ii];
        vpriv[ii]->device = &(pdev->dev);
	    vpriv[ii]->plat = plat_dat;

        ether_setup(vndev[ii]);

        vndev[ii]->netdev_ops = &v_stmmac_netdev_ops;
               
        /**get phy addr*/
#ifdef CONFIG_RTL8363_CPUPORT
        if(virt_ndev[ii].port == GXE_SVIP_PORT)
#else
        if(virt_ndev[ii].vlan_id == SVIP_VLANID)
#endif
            vpriv[ii]->phy_addr = -1;
        else
            vpriv[ii]->phy_addr = get_free_phydev();
        
        stmmac_set_ethtool_ops(vndev[ii]);

        vndev[ii]->features |= NETIF_F_SG | NETIF_F_HIGHDMA |
            NETIF_F_IP_CSUM | NETIF_F_IPV6_CSUM;
        vndev[ii]->watchdog_timeo = msecs_to_jiffies(watchdog);
#ifdef STMMAC_VLAN_TAG_USED
     /* Both mac100 and gmac support receive VLAN tag detection */
        vndev[ii]->features |= NETIF_F_HW_VLAN_RX;
#endif
        vpriv[ii]->msg_enable = netif_msg_init(debug, default_msg_level);

        if (flow_ctrl)
            vpriv[ii]->flow_ctrl = FLOW_AUTO;    /* RX/TX pause on */

        vpriv[ii]->pause = pause;

        /* Get the MAC address */
        if (real_ndev){
            rdev_priv = netdev_priv(real_ndev);
            vpriv[ii]->ioaddr = rdev_priv->ioaddr;
            
        /* MAC HW revice detection */
            vpriv[ii]->hw = rdev_priv->hw;
            rdev_priv->hw->mac->get_umac_addr((void __iomem *) real_ndev->base_addr,
                         vndev[ii]->dev_addr, 0);
            
            if (!is_valid_ether_addr(vndev[ii]->dev_addr))
                pr_warning("\tno valid MAC address;"
                    "please, use ifconfig or nwhwconfig!\n");
        }
    
        spin_lock_init(&vpriv[ii]->lock);

        ret = register_netdev(vndev[ii]);
        if (ret) {
            pr_err("%s: ERROR %i registering the device\n",
                   __func__, ret);
            free_netdev(vndev[ii]);      
            return -ENODEV;
        }
 
        pr_info("\t%s - (dev. name: %s - id: %d\n"
             "\tIO base addr: 0x%p)\n", vndev[ii]->name, pdev->name,
             pdev->id, vpriv[ii]->ioaddr);

        DBG(probe, DEBUG, "%s: Scatter/Gather: %s - HW checksums: %s\n",
            vndev[ii]->name, (vndev[ii]->features & NETIF_F_SG) ? "on" : "off",
            (vndev[ii]->features & NETIF_F_IP_CSUM) ? "on" : "off");
    }
    return ret;
}

static void del_virt_ndev(void)
{
    int ii;
    for(ii=0;ii<NUM_OF_VDEV;ii++){
        if(virt_ndev[ii].ndev) {
            netif_carrier_off(virt_ndev[ii].ndev);
            unregister_netdev(virt_ndev[ii].ndev);
            free_netdev(virt_ndev[ii].ndev);
            virt_ndev[ii].ndev = NULL;
        }
    }
    g_mdio_priv.phy_inused = 0;
}
#endif

/**
 * stmmac_dvr_probe
 * @pdev: platform device pointer
 * Description: the driver is initialized through platform_device.
 */
static int stmmac_dvr_probe(struct platform_device *pdev)
{
	int ret = 0;
	struct resource *res;
	void __iomem *addr = NULL;
	struct net_device *ndev = NULL;
	struct stmmac_priv *priv = NULL;
	struct plat_stmmacenet_data *plat_dat;
    
    /** FIXME: for bug[15135] we reset rtl8363 in uboot, and take out of 
        reset in kernel board init, and we must sleep for a short time , 
        if not the dwmac1000_dma_init will return BUSY.
    */
    msleep(500);
    
	pr_info("STMMAC driver:\n\tplatform registration... ");
	res = platform_get_resource(pdev, IORESOURCE_MEM, 0);
	if (!res)
		return -ENODEV;
	pr_info("\tdone!\n");

	if (!request_mem_region(res->start, resource_size(res),
				pdev->name)) {
		pr_err("%s: ERROR: memory allocation failed"
		       "cannot get the I/O addr 0x%x\n",
		       __func__, (unsigned int)res->start);
		return -EBUSY;
	}

	addr = ioremap(res->start, resource_size(res));
	if (!addr) {
		pr_err("%s: ERROR: memory mapping failed\n", __func__);
		ret = -ENOMEM;
		goto out_release_region;
	}

	//ndev = alloc_etherdev(sizeof(struct stmmac_priv));
	ndev = alloc_netdev_mqs(sizeof(struct stmmac_priv), REAL_NDEV_NAME, ether_setup, 1, 1);

	if (!ndev) {
		pr_err("%s: ERROR: allocating the device\n", __func__);
		ret = -ENOMEM;
		goto out_unmap;
	}

	SET_NETDEV_DEV(ndev, &pdev->dev);

	/* Get the MAC information */
	ndev->irq = platform_get_irq_byname(pdev, "macirq");
	if (ndev->irq == -ENXIO) {
		pr_err("%s: ERROR: MAC IRQ configuration "
		       "information not found\n", __func__);
		ret = -ENXIO;
		goto out_free_ndev;
	}

	priv = netdev_priv(ndev);
	priv->device = &(pdev->dev);
	priv->dev = ndev;
	plat_dat = pdev->dev.platform_data;

	priv->plat = plat_dat;

	priv->ioaddr = addr;

	/* PMT module is not integrated in all the MAC devices. */
	if (plat_dat->pmt) {
		pr_info("\tPMT module supported\n");
		device_set_wakeup_capable(&pdev->dev, 1);
	}

	platform_set_drvdata(pdev, ndev);

	/* Set the I/O base addr */
	ndev->base_addr = (unsigned long)addr;

	/* Custom initialisation */
	if (priv->plat->init) {
		ret = priv->plat->init(pdev);
		if (unlikely(ret))
			goto out_free_ndev;
	}

	/* MAC HW revice detection */
	ret = stmmac_mac_device_setup(ndev);
	if (ret < 0)
		goto out_plat_exit;

	/* Network Device Registration */
	ret = stmmac_probe(ndev);
	if (ret < 0)
		goto out_plat_exit;
    
#ifndef USE_VIRTUAL_ETH
	/* associate a PHY - it is provided by another platform bus */
	if (!driver_for_each_device
	    (&(stmmacphy_driver.driver), NULL, (void *)priv,
	     stmmac_associate_phy)) {
		pr_err("No PHY device is associated with this MAC!\n");
		ret = -ENODEV;
		goto out_unregister;
	}
	/* MDIO bus Registration */
	pr_debug("\tMDIO bus (id: %d)...", priv->plat->bus_id);
	ret = stmmac_mdio_register(ndev);
	if (ret < 0)
		goto out_unregister;
#endif   
	pr_info("\t%s - (dev. name: %s - id: %d, IRQ #%d\n"
	       "\tIO base addr: 0x%p)\n", ndev->name, pdev->name,
	       pdev->id, ndev->irq, addr);

#ifdef USE_VIRTUAL_ETH 
    /* MDIO bus Registration */
    /* associate a PHY - it is provided by another platform bus */
    g_mdio_priv.has_nphys = 0;

    if(driver_for_each_device(&(stmmacphy_driver.driver), NULL,
        (void *)&g_mdio_priv,
         stmmac_veth_associate_phy)){
         pr_warn("No Enought PHY for stmmacphy platform device ");
    };
    if(0 == g_mdio_priv.has_nphys){
        pr_err("No PHY device is associated with this MAC!\n");
        ret = -ENODEV;
        goto out_unregister;
    }
    g_mdio_priv.ioaddr = priv->ioaddr;
    g_mdio_priv.plat = priv->plat;
    g_mdio_priv.hw = priv->hw;
    g_mdio_priv.device = priv->device;
   
    pr_debug("\tMDIO bus (id: %d)...", priv->plat->bus_id);
    ret = stmmac_mdio_register(&g_mdio_priv);
    if (ret < 0)
        goto out_unregister;

    real_ndev = ndev;
    ret = add_virt_ndev(pdev);
    if(ret<0){
        stmmac_mdio_unregister(&g_mdio_priv);
        goto out_unregister;
    }

    ret = gxe_rtl8363_install_proc_entry();
    if(ret<0){
        pr_debug("gxe rtl8363 install proc entry error!");
    }

#endif
    
	pr_debug("registered!\n");
	return 0;

out_unregister:
	unregister_netdev(ndev);
out_plat_exit:
	if (priv->plat->exit)
		priv->plat->exit(pdev);
out_free_ndev:
	free_netdev(ndev);
	platform_set_drvdata(pdev, NULL);
out_unmap:
	iounmap(addr);
out_release_region:
	release_mem_region(res->start, resource_size(res));

	return ret;
}

/**
 * stmmac_dvr_remove
 * @pdev: platform device pointer
 * Description: this function resets the TX/RX processes, disables the MAC RX/TX
 * changes the link status, releases the DMA descriptor rings,
 * unregisters the MDIO bus and unmaps the allocated memory.
 */
static int stmmac_dvr_remove(struct platform_device *pdev)
{
	struct net_device *ndev = platform_get_drvdata(pdev);
	struct stmmac_priv *priv = netdev_priv(ndev);
	struct resource *res;

	pr_info("%s:\n\tremoving driver\n", __func__);
#ifdef USE_VIRTUAL_ETH  
    gxe_rtl8363_remove_proc_entry();
    del_virt_ndev();
#endif
	priv->hw->dma->stop_rx(priv->ioaddr);
	priv->hw->dma->stop_tx(priv->ioaddr);

	stmmac_disable_mac(priv->ioaddr);

	netif_carrier_off(ndev);
    
#ifdef USE_VIRTUAL_ETH 
	stmmac_mdio_unregister(&g_mdio_priv);
#else
	stmmac_mdio_unregister(ndev);
#endif

	if (priv->plat->exit)
		priv->plat->exit(pdev);

	platform_set_drvdata(pdev, NULL);
	unregister_netdev(ndev);
    
	iounmap((void *)priv->ioaddr);
	res = platform_get_resource(pdev, IORESOURCE_MEM, 0);
	release_mem_region(res->start, resource_size(res));
    
	free_netdev(ndev);
    
#ifdef USE_VIRTUAL_ETH     
    real_ndev = NULL;
#endif
	return 0;
}

#ifdef CONFIG_PM
static int stmmac_suspend(struct device *dev)
{
	struct net_device *ndev = dev_get_drvdata(dev);
	struct stmmac_priv *priv = netdev_priv(ndev);
	int dis_ic = 0;

	if (!ndev || !netif_running(ndev))
		return 0;

	spin_lock(&priv->lock);

	netif_device_detach(ndev);
	netif_stop_queue(ndev);
	if (priv->phydev)
		phy_stop(priv->phydev);

#ifdef CONFIG_STMMAC_TIMER
	priv->tm->timer_stop();
	if (likely(priv->tm->enable))
		dis_ic = 1;
#endif
	napi_disable(&priv->napi);

	/* Stop TX/RX DMA */
	priv->hw->dma->stop_tx(priv->ioaddr);
	priv->hw->dma->stop_rx(priv->ioaddr);
	/* Clear the Rx/Tx descriptors */
	priv->hw->desc->init_rx_desc(priv->dma_rx, priv->dma_rx_size,
				     dis_ic);
	priv->hw->desc->init_tx_desc(priv->dma_tx, priv->dma_tx_size);

	/* Enable Power down mode by programming the PMT regs */
	if (device_may_wakeup(priv->device))
		priv->hw->mac->pmt(priv->ioaddr, priv->wolopts);
	else
		stmmac_disable_mac(priv->ioaddr);

	spin_unlock(&priv->lock);
	return 0;
}

static int stmmac_resume(struct device *dev)
{
	struct net_device *ndev = dev_get_drvdata(dev);
	struct stmmac_priv *priv = netdev_priv(ndev);

	if (!netif_running(ndev))
		return 0;

	spin_lock(&priv->lock);

	/* Power Down bit, into the PM register, is cleared
	 * automatically as soon as a magic packet or a Wake-up frame
	 * is received. Anyway, it's better to manually clear
	 * this bit because it can generate problems while resuming
	 * from another devices (e.g. serial console). */
	if (device_may_wakeup(priv->device))
		priv->hw->mac->pmt(priv->ioaddr, 0);

	netif_device_attach(ndev);

	/* Enable the MAC and DMA */
	stmmac_enable_mac(priv->ioaddr);
	priv->hw->dma->start_tx(priv->ioaddr);
	priv->hw->dma->start_rx(priv->ioaddr);

#ifdef CONFIG_STMMAC_TIMER
	if (likely(priv->tm->enable))
		priv->tm->timer_start(tmrate);
#endif
	napi_enable(&priv->napi);

	if (priv->phydev)
		phy_start(priv->phydev);

	netif_start_queue(ndev);

	spin_unlock(&priv->lock);
	return 0;
}

static int stmmac_freeze(struct device *dev)
{
	struct net_device *ndev = dev_get_drvdata(dev);

	if (!ndev || !netif_running(ndev))
		return 0;

	return stmmac_release(ndev);
}

static int stmmac_restore(struct device *dev)
{
	struct net_device *ndev = dev_get_drvdata(dev);

	if (!ndev || !netif_running(ndev))
		return 0;

	return stmmac_open(ndev);
}

static const struct dev_pm_ops stmmac_pm_ops = {
	.suspend = stmmac_suspend,
	.resume = stmmac_resume,
	.freeze = stmmac_freeze,
	.thaw = stmmac_restore,
	.restore = stmmac_restore,
};
#else
static const struct dev_pm_ops stmmac_pm_ops;
#endif /* CONFIG_PM */

static struct platform_driver stmmac_driver = {
	.probe = stmmac_dvr_probe,
	.remove = stmmac_dvr_remove,
	.driver = {
		.name = STMMAC_RESOURCE_NAME,
		.owner = THIS_MODULE,
		.pm = &stmmac_pm_ops,
	},
};

/**
 * stmmac_init_module - Entry point for the driver
 * Description: This function is the entry point for the driver.
 */
static int __init stmmac_init_module(void)
{
	int ret;

	if (platform_driver_register(&stmmacphy_driver)) {
		pr_err("No PHY devices registered!\n");
		return -ENODEV;
	}

	ret = platform_driver_register(&stmmac_driver);
	return ret;
}

/**
 * stmmac_cleanup_module - Cleanup routine for the driver
 * Description: This function is the cleanup routine for the driver.
 */
static void __exit stmmac_cleanup_module(void)
{
	platform_driver_unregister(&stmmacphy_driver);
	platform_driver_unregister(&stmmac_driver);
}

#ifndef MODULE
static int __init stmmac_cmdline_opt(char *str)
{
    char *opt;
    int ret = 0;

    if (!str || !*str)
        return -EINVAL;
    while ((opt = strsep(&str, ",")) != NULL) {
        if (!strncmp(opt, "debug:", 6))
            ret = strict_strtoul(opt + 6, 0, (unsigned long *)&debug);
        else if (!strncmp(opt, "phyaddr:", 8))
            ret = strict_strtoul(opt + 8, 0, (unsigned long *)&phyaddr);
        else if (!strncmp(opt, "dma_txsize:", 11)) 
            ret = strict_strtoul(opt + 11, 0,
                       (unsigned long *)&dma_txsize);
        else if (!strncmp(opt, "dma_rxsize:", 11)) 
            ret = strict_strtoul(opt + 11, 0,
                       (unsigned long *)&dma_rxsize);
        else if (!strncmp(opt, "buf_sz:", 7))
            ret = strict_strtoul(opt + 7, 0, (unsigned long *)&buf_sz);
        else if (!strncmp(opt, "tc:", 3))
            ret = strict_strtoul(opt + 3, 0, (unsigned long *)&tc);
        else if (!strncmp(opt, "watchdog:", 9))
            ret = strict_strtoul(opt + 9, 0, (unsigned long *)&watchdog);
        else if (!strncmp(opt, "flow_ctrl:", 10)) 
            ret = strict_strtoul(opt + 10, 0,
                       (unsigned long *)&flow_ctrl);
        else if (!strncmp(opt, "pause:", 6))
            ret = strict_strtoul(opt + 6, 0, (unsigned long *)&pause);
#ifdef CONFIG_STMMAC_TIMER
        else if (!strncmp(opt, "tmrate:", 7))
            ret = strict_strtoul(opt + 7, 0, (unsigned long *)&tmrate);
#endif
    }    
    return ret;
}

__setup("stmmaceth=", stmmac_cmdline_opt);
#endif

module_init(stmmac_init_module);
module_exit(stmmac_cleanup_module);

MODULE_DESCRIPTION("STMMAC 10/100/1000 Ethernet driver");
MODULE_AUTHOR("Giuseppe Cavallaro <peppe.cavallaro@st.com>");
MODULE_LICENSE("GPL");
